import asyncio

from pyensign.ensign import Ensign


class Subscriber(Ensign):
    """
    A Subscriber is a high-level abstraction for consuming events from an Ensign topic.
    Subscribers consume events from a topic and pass them to a callback function.
    """

    def __init__(self, *topics, **kwargs):
        """
        Create a Subscriber to consume events from a set of topics. Once initialized,
        call run() to start consuming events from the topics.

        Parameters
        ----------
        *topics : iterable of str
            The topic names or IDs to consume events from.

        **kwargs
            Credentials and configuration for the Ensign client, see the Ensign class
            for more details.

        Raises
        ------
        ValueError
            If no topics are provided.
        """

        super().__init__(**kwargs)

        if len(topics) == 0:
            raise ValueError("no topics provided")
        self.subscribed_topics = topics

    async def on_event(self, event):
        """
        Implement this method to process events as they are received. By default, this
        method simply ACKs the events back to the server.
        """

        await event.ack()

    def run(self, **kwargs):
        """
        Consume events from the topics and invoke the callback for each event. This
        method blocks but may raise an EnsignError if the subscriber encounters an
        error communicating with Ensign (e.g. a connection timeout or topic not found error).

        Parameters
        ----------

        **kwargs
            Keyword arguments to pass to the subscribe() method, e.g. enSQL queries or
            consumer group IDs.
        """

        exception = asyncio.run(self.consume(), **kwargs)
        if exception:
            raise exception

    async def consume(self, **kwargs):
        """
        Consume events from the subscriber and execute the callback for each event.
        """

        async for event in self.subscribe(self.subscribed_topics, **kwargs):
            await self.on_event(event)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/v1beta1/query.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x17\x61pi/v1beta1/query.proto\x12\x0e\x65nsign.v1beta1"A\n\x05Query\x12\r\n\x05query\x18\x01 \x01(\t\x12)\n\x06params\x18\x02 \x03(\x0b\x32\x19.ensign.v1beta1.Parameter"c\n\tParameter\x12\x0b\n\x01i\x18\x01 \x01(\x12H\x00\x12\x0b\n\x01\x64\x18\x02 \x01(\x01H\x00\x12\x0b\n\x01\x62\x18\x03 \x01(\x08H\x00\x12\x0b\n\x01y\x18\x04 \x01(\x0cH\x00\x12\x0b\n\x01s\x18\x05 \x01(\tH\x00\x12\x0c\n\x04name\x18\x06 \x01(\tB\x07\n\x05value"\x12\n\x10QueryExplanationb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "api.v1beta1.query_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _QUERY._serialized_start = 43
    _QUERY._serialized_end = 108
    _PARAMETER._serialized_start = 110
    _PARAMETER._serialized_end = 209
    _QUERYEXPLANATION._serialized_start = 211
    _QUERYEXPLANATION._serialized_end = 229
# @@protoc_insertion_point(module_scope)



import asyncio
import inspect

from pyensign.ensign import Ensign


class Publisher(Ensign):
    """
    A Publisher is a high-level abstraction for publishing events to Ensign. Publishers
    consume an iterator of events and publish them to an Ensign topic.
    """

    def __init__(self, topic, **kwargs):
        """
        Create a Publisher to publish events to a topic. Once initialized, call run()
        to publish events to the topic.

        Parameters
        ----------
        topic : str
            The topic name or ID to publish events to.

        **kwargs
            Credentials and configuration for the Ensign client, see the Ensign class
            for more details.

        Raises
        ------
        ValueError
            If a topic is not provided.
        """

        super().__init__(**kwargs)

        if not topic:
            raise ValueError("topic is required")
        self.topic = topic

    def run(self, event_iterator, **kwargs):
        """
        Publish all the events from the iterator to the topic. This method blocks until
        the iterator is exhausted.

        Parameters
        ----------
        event_iterator : Iterator or AsyncIterator
            An iterator of events to publish to the topic.

        **kwargs
            Additional keyword arguments to pass to the publish() method.
        """

        exception = asyncio.run(self.publish_from(event_iterator, **kwargs))
        if exception:
            raise exception

    async def publish_from(self, event_iterator, **kwargs):
        """
        Publish all the events from the iterator to the topic. This method blocks until
        all the events have been published.

        Parameters
        ----------
        event_iterator : Iterator or AsyncIterator
            An iterator of events to publish to the topic.

        **kwargs
            Additional keyword arguments to pass to the publish() method.
        """

        if inspect.isasyncgen(event_iterator):
            async for event in event_iterator:
                try:
                    await self.publish(self.topic, event, **kwargs)
                except Exception as e:
                    return e
        else:
            for event in event_iterator:
                try:
                    await self.publish(self.topic, event, **kwargs)
                except Exception as e:
                    return e



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/v1beta1/event.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pyensign.region.v1beta1 import region_pb2 as region_dot_v1beta1_dot_region__pb2
from pyensign.mimetype.v1beta1 import (
    mimetype_pb2 as mimetype_dot_v1beta1_dot_mimetype__pb2,
)
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x17\x61pi/v1beta1/event.proto\x12\x0e\x65nsign.v1beta1\x1a\x1bregion/v1beta1/region.proto\x1a\x1fmimetype/v1beta1/mimetype.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xef\x02\n\x0c\x45ventWrapper\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\x10\n\x08topic_id\x18\x02 \x01(\x0c\x12\x0e\n\x06offset\x18\x03 \x01(\x04\x12\r\n\x05\x65poch\x18\x04 \x01(\x04\x12&\n\x06region\x18\x05 \x01(\x0e\x32\x16.region.v1beta1.Region\x12,\n\tpublisher\x18\x06 \x01(\x0b\x32\x19.ensign.v1beta1.Publisher\x12\x0b\n\x03key\x18\x07 \x01(\x0c\x12\r\n\x05shard\x18\x08 \x01(\x04\x12\r\n\x05\x65vent\x18\t \x01(\x0c\x12.\n\nencryption\x18\n \x01(\x0b\x32\x1a.ensign.v1beta1.Encryption\x12\x30\n\x0b\x63ompression\x18\x0b \x01(\x0b\x32\x1b.ensign.v1beta1.Compression\x12-\n\tcommitted\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08local_id\x18\x10 \x01(\x0c"\xf8\x01\n\x05\x45vent\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x35\n\x08metadata\x18\x03 \x03(\x0b\x32#.ensign.v1beta1.Event.MetadataEntry\x12(\n\x08mimetype\x18\x04 \x01(\x0e\x32\x16.mimetype.v1beta1.MIME\x12"\n\x04type\x18\x05 \x01(\x0b\x32\x14.ensign.v1beta1.Type\x12+\n\x07\x63reated\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xe3\x07\n\x0e\x45ventContainer\x12\x10\n\x08topic_id\x18\x01 \x01(\x0c\x12\x14\n\x0cstart_offset\x18\x02 \x01(\x04\x12\x12\n\nend_offset\x18\x03 \x01(\x04\x12:\n\x06\x65pochs\x18\x04 \x03(\x0b\x32*.ensign.v1beta1.EventContainer.EpochsEntry\x12\x0e\n\x06\x65vents\x18\x05 \x01(\x0c\x12.\n\nencryption\x18\x06 \x01(\x0b\x32\x1a.ensign.v1beta1.Encryption\x12\x30\n\x0b\x63ompression\x18\x07 \x01(\x0b\x32\x1b.ensign.v1beta1.Compression\x12\'\n\x07regions\x18\x08 \x03(\x0e\x32\x16.region.v1beta1.Region\x12\x45\n\x0cregion_index\x18\t \x03(\x0b\x32/.ensign.v1beta1.EventContainer.RegionIndexEntry\x12-\n\npublishers\x18\n \x03(\x0b\x32\x19.ensign.v1beta1.Publisher\x12K\n\x0fpublisher_index\x18\x0b \x03(\x0b\x32\x32.ensign.v1beta1.EventContainer.PublisherIndexEntry\x12\x0c\n\x04keys\x18\x0c \x03(\x0c\x12?\n\tkey_index\x18\r \x03(\x0b\x32,.ensign.v1beta1.EventContainer.KeyIndexEntry\x12\x0e\n\x06shards\x18\x0e \x03(\x04\x12\x43\n\x0bshard_index\x18\x0f \x03(\x0b\x32..ensign.v1beta1.EventContainer.ShardIndexEntry\x12+\n\x07\x63reated\x18\x1f \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08modified\x18  \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a-\n\x0b\x45pochsEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a\x32\n\x10RegionIndexEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a\x35\n\x13PublisherIndexEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a/\n\rKeyIndexEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a\x31\n\x0fShardIndexEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01"Y\n\x04Type\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rmajor_version\x18\x02 \x01(\r\x12\x15\n\rminor_version\x18\x03 \x01(\r\x12\x15\n\rpatch_version\x18\x04 \x01(\r"\xa0\x03\n\nEncryption\x12\x15\n\rpublic_key_id\x18\x01 \x01(\t\x12\x16\n\x0e\x65ncryption_key\x18\x02 \x01(\x0c\x12\x13\n\x0bhmac_secret\x18\x03 \x01(\x0c\x12\x11\n\tsignature\x18\x04 \x01(\x0c\x12?\n\x11sealing_algorithm\x18\x05 \x01(\x0e\x32$.ensign.v1beta1.Encryption.Algorithm\x12\x42\n\x14\x65ncryption_algorithm\x18\x06 \x01(\x0e\x32$.ensign.v1beta1.Encryption.Algorithm\x12\x41\n\x13signature_algorithm\x18\x07 \x01(\x0e\x32$.ensign.v1beta1.Encryption.Algorithm"s\n\tAlgorithm\x12\r\n\tPLAINTEXT\x10\x00\x12\x0e\n\nAES256_GCM\x10n\x12\x0e\n\nAES192_GCM\x10x\x12\x0f\n\nAES128_GCM\x10\x82\x01\x12\x10\n\x0bHMAC_SHA256\x10\xb6\x02\x12\x14\n\x0fRSA_OAEP_SHA512\x10\xfe\x03"\x9e\x01\n\x0b\x43ompression\x12\x38\n\talgorithm\x18\x01 \x01(\x0e\x32%.ensign.v1beta1.Compression.Algorithm\x12\r\n\x05level\x18\x02 \x01(\x03"F\n\tAlgorithm\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04GZIP\x10\x01\x12\x0c\n\x08\x43OMPRESS\x10\x02\x12\x0b\n\x07\x44\x45\x46LATE\x10\x03\x12\n\n\x06\x42ROTLI\x10\x04"X\n\tPublisher\x12\x14\n\x0cpublisher_id\x18\x01 \x01(\t\x12\x0e\n\x06ipaddr\x18\x02 \x01(\t\x12\x11\n\tclient_id\x18\x03 \x01(\t\x12\x12\n\nuser_agent\x18\x04 \x01(\tb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "api.v1beta1.event_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _EVENT_METADATAENTRY._options = None
    _EVENT_METADATAENTRY._serialized_options = b"8\001"
    _EVENTCONTAINER_EPOCHSENTRY._options = None
    _EVENTCONTAINER_EPOCHSENTRY._serialized_options = b"8\001"
    _EVENTCONTAINER_REGIONINDEXENTRY._options = None
    _EVENTCONTAINER_REGIONINDEXENTRY._serialized_options = b"8\001"
    _EVENTCONTAINER_PUBLISHERINDEXENTRY._options = None
    _EVENTCONTAINER_PUBLISHERINDEXENTRY._serialized_options = b"8\001"
    _EVENTCONTAINER_KEYINDEXENTRY._options = None
    _EVENTCONTAINER_KEYINDEXENTRY._serialized_options = b"8\001"
    _EVENTCONTAINER_SHARDINDEXENTRY._options = None
    _EVENTCONTAINER_SHARDINDEXENTRY._serialized_options = b"8\001"
    _EVENTWRAPPER._serialized_start = 139
    _EVENTWRAPPER._serialized_end = 506
    _EVENT._serialized_start = 509
    _EVENT._serialized_end = 757
    _EVENT_METADATAENTRY._serialized_start = 710
    _EVENT_METADATAENTRY._serialized_end = 757
    _EVENTCONTAINER._serialized_start = 760
    _EVENTCONTAINER._serialized_end = 1755
    _EVENTCONTAINER_EPOCHSENTRY._serialized_start = 1503
    _EVENTCONTAINER_EPOCHSENTRY._serialized_end = 1548
    _EVENTCONTAINER_REGIONINDEXENTRY._serialized_start = 1550
    _EVENTCONTAINER_REGIONINDEXENTRY._serialized_end = 1600
    _EVENTCONTAINER_PUBLISHERINDEXENTRY._serialized_start = 1602
    _EVENTCONTAINER_PUBLISHERINDEXENTRY._serialized_end = 1655
    _EVENTCONTAINER_KEYINDEXENTRY._serialized_start = 1657
    _EVENTCONTAINER_KEYINDEXENTRY._serialized_end = 1704
    _EVENTCONTAINER_SHARDINDEXENTRY._serialized_start = 1706
    _EVENTCONTAINER_SHARDINDEXENTRY._serialized_end = 1755
    _TYPE._serialized_start = 1757
    _TYPE._serialized_end = 1846
    _ENCRYPTION._serialized_start = 1849
    _ENCRYPTION._serialized_end = 2265
    _ENCRYPTION_ALGORITHM._serialized_start = 2150
    _ENCRYPTION_ALGORITHM._serialized_end = 2265
    _COMPRESSION._serialized_start = 2268
    _COMPRESSION._serialized_end = 2426
    _COMPRESSION_ALGORITHM._serialized_start = 2356
    _COMPRESSION_ALGORITHM._serialized_end = 2426
    _PUBLISHER._serialized_start = 2428
    _PUBLISHER._serialized_end = 2516
# @@protoc_insertion_point(module_scope)



from pyensign.exceptions import CacheMissError


class Cache:
    """
    Cache is a simple cache that maps keys to values to avoid repeated RPC calls to
    Ensign. This is not thread-safe.
    TODO: Implement max size and evictions.
    """

    def __init__(self, read_only=False):
        """
        Create a new cache.

        Parameters
        ----------
        read_only : bool (default: False)
            If True, the cache will not be writeable. This creates a disabled cache
            where get() will always raise a CacheMissError, exists() will always return
            False, and add() is a no-op.
        """
        self._index = {}
        self._read_only = read_only

    def get(self, key):
        """
        Get a value by name, an exception is raised if the key does not exist in the
        cache.
        """
        try:
            return self._index[key]
        except KeyError as e:
            raise CacheMissError(key) from e

    def add(self, key, value):
        """
        Add a value by key to the cache, overwriting the existing key.
        """
        if not self._read_only:
            self._index[key] = value

    def exists(self, key):
        """
        Returns True if the key exists in the cache.
        """
        return key in self._index

    def clear(self):
        """
        Reset the cache, deleting all keys.
        """
        if not self._read_only:
            self._index.clear()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/v1beta1/topic.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pyensign.api.v1beta1 import event_pb2 as api_dot_v1beta1_dot_event__pb2
from pyensign.region.v1beta1 import region_pb2 as region_dot_v1beta1_dot_region__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from pyensign.mimetype.v1beta1 import (
    mimetype_pb2 as mimetype_dot_v1beta1_dot_mimetype__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x17\x61pi/v1beta1/topic.proto\x12\x0e\x65nsign.v1beta1\x1a\x17\x61pi/v1beta1/event.proto\x1a\x1bregion/v1beta1/region.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1fmimetype/v1beta1/mimetype.proto"\x96\x02\n\x05Topic\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\x12\n\nproject_id\x18\x02 \x01(\x0c\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x10\n\x08readonly\x18\x04 \x01(\x08\x12\x0e\n\x06offset\x18\x05 \x01(\x04\x12\x0e\n\x06shards\x18\x06 \x01(\r\x12-\n\nplacements\x18\x0c \x03(\x0b\x32\x19.ensign.v1beta1.Placement\x12#\n\x05types\x18\r \x03(\x0b\x32\x14.ensign.v1beta1.Type\x12+\n\x07\x63reated\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08modified\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp"?\n\tTopicName\x12\x10\n\x08topic_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t"\xe3\x01\n\tTopicInfo\x12\x10\n\x08topic_id\x18\x01 \x01(\x0c\x12\x12\n\nproject_id\x18\x02 \x01(\x0c\x12\x17\n\x0f\x65vent_offset_id\x18\x03 \x01(\x0c\x12\x0e\n\x06\x65vents\x18\x07 \x01(\x04\x12\x12\n\nduplicates\x18\x08 \x01(\x04\x12\x17\n\x0f\x64\x61ta_size_bytes\x18\t \x01(\x04\x12,\n\x05types\x18\x0e \x03(\x0b\x32\x1d.ensign.v1beta1.EventTypeInfo\x12,\n\x08modified\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp"L\n\nTopicsPage\x12%\n\x06topics\x18\x01 \x03(\x0b\x32\x15.ensign.v1beta1.Topic\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t"Y\n\x0eTopicNamesPage\x12.\n\x0btopic_names\x18\x01 \x03(\x0b\x32\x19.ensign.v1beta1.TopicName\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t"~\n\x08TopicMod\x12\n\n\x02id\x18\x01 \x01(\t\x12\x35\n\toperation\x18\x02 \x01(\x0e\x32".ensign.v1beta1.TopicMod.Operation"/\n\tOperation\x12\x08\n\x04NOOP\x10\x00\x12\x0b\n\x07\x41RCHIVE\x10\x01\x12\x0b\n\x07\x44\x45STROY\x10\x02"\x85\x01\n\x0eTopicTombstone\x12\n\n\x02id\x18\x01 \x01(\t\x12\x34\n\x05state\x18\x02 \x01(\x0e\x32%.ensign.v1beta1.TopicTombstone.Status"1\n\x06Status\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08READONLY\x10\x01\x12\x0c\n\x08\x44\x45LETING\x10\x02"0\n\x0fTopicExistsInfo\x12\r\n\x05query\x18\x01 \x01(\t\x12\x0e\n\x06\x65xists\x18\x02 \x01(\x08"\x9c\x01\n\tPlacement\x12\r\n\x05\x65poch\x18\x01 \x01(\x04\x12\x32\n\x08sharding\x18\x02 \x01(\x0e\x32 .ensign.v1beta1.ShardingStrategy\x12\'\n\x07regions\x18\x03 \x03(\x0e\x32\x16.region.v1beta1.Region\x12#\n\x05nodes\x18\x04 \x03(\x0b\x32\x14.ensign.v1beta1.Node"x\n\x04Node\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08hostname\x18\x02 \x01(\t\x12\x0e\n\x06quorum\x18\x03 \x01(\x04\x12\r\n\x05shard\x18\x04 \x01(\x04\x12&\n\x06region\x18\x05 \x01(\x0e\x32\x16.region.v1beta1.Region\x12\x0b\n\x03url\x18\x06 \x01(\t"\xc8\x01\n\rEventTypeInfo\x12"\n\x04type\x18\x01 \x01(\x0b\x32\x14.ensign.v1beta1.Type\x12(\n\x08mimetype\x18\x02 \x01(\x0e\x32\x16.mimetype.v1beta1.MIME\x12\x0e\n\x06\x65vents\x18\n \x01(\x04\x12\x12\n\nduplicates\x18\x0b \x01(\x04\x12\x17\n\x0f\x64\x61ta_size_bytes\x18\x0c \x01(\x04\x12,\n\x08modified\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp*m\n\x10ShardingStrategy\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0f\n\x0bNO_SHARDING\x10\x01\x12\x17\n\x13\x43ONSISTENT_KEY_HASH\x10\x02\x12\n\n\x06RANDOM\x10\x03\x12\x16\n\x12PUBLISHER_ORDERING\x10\x04\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "api.v1beta1.topic_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SHARDINGSTRATEGY._serialized_start = 1706
    _SHARDINGSTRATEGY._serialized_end = 1815
    _TOPIC._serialized_start = 164
    _TOPIC._serialized_end = 442
    _TOPICNAME._serialized_start = 444
    _TOPICNAME._serialized_end = 507
    _TOPICINFO._serialized_start = 510
    _TOPICINFO._serialized_end = 737
    _TOPICSPAGE._serialized_start = 739
    _TOPICSPAGE._serialized_end = 815
    _TOPICNAMESPAGE._serialized_start = 817
    _TOPICNAMESPAGE._serialized_end = 906
    _TOPICMOD._serialized_start = 908
    _TOPICMOD._serialized_end = 1034
    _TOPICMOD_OPERATION._serialized_start = 987
    _TOPICMOD_OPERATION._serialized_end = 1034
    _TOPICTOMBSTONE._serialized_start = 1037
    _TOPICTOMBSTONE._serialized_end = 1170
    _TOPICTOMBSTONE_STATUS._serialized_start = 1121
    _TOPICTOMBSTONE_STATUS._serialized_end = 1170
    _TOPICEXISTSINFO._serialized_start = 1172
    _TOPICEXISTSINFO._serialized_end = 1220
    _PLACEMENT._serialized_start = 1223
    _PLACEMENT._serialized_end = 1379
    _NODE._serialized_start = 1381
    _NODE._serialized_end = 1501
    _EVENTTYPEINFO._serialized_start = 1504
    _EVENTTYPEINFO._serialized_end = 1704
# @@protoc_insertion_point(module_scope)



import grpc
import asyncio
from ulid import ULID
from grpc import aio
from datetime import timedelta

from pyensign.events import from_proto
from pyensign.version import user_agent
from pyensign.utils.tasks import WorkerPool
from pyensign.api.v1beta1.event import unwrap
from pyensign.exceptions import catch_rpc_error
from pyensign.api.v1beta1 import ensign_pb2_grpc
from pyensign.api.v1beta1 import topic_pb2, ensign_pb2, query_pb2
from pyensign.auth.interceptor import (
    MetadataUnaryInterceptor,
    MetadataUnaryStreamInterceptor,
    MetadataStreamInterceptor,
)
from pyensign.stream import Publisher, Subscriber
from pyensign.exceptions import (
    QueryNoRows,
    EnsignClientClosingError,
)


class Connection:
    """
    Connection defines a gRPC connection to an Ensign server.
    """

    def __init__(self, addrport, insecure=False, auth=None):
        """
        Connect to an Ensign server.

        Parameters
        ----------
        addrport : str
            The address:port of the Ensign server (e.g. "localhost:5356")
        insecure : bool
            Set to True to use an insecure connection. This is only useful for testing
            against a local Ensign server and overrides the auth parameter.
        auth : grpc.AuthMetadataPlugin or None
            Plugin that provides an access token for per-request authentication.
        """

        addrParts = addrport.split(":", 2)
        if len(addrParts) != 2 or not addrParts[0] or not addrParts[1]:
            raise ValueError("Invalid address:port format")
        self.addrport = addrport

        self.insecure = insecure
        self.auth = auth

    def create_channel(self):
        """
        Create a new gRPC channel from the connection parameters. The caller is
        responsible for closing the channel.
        """

        # Create interceptors for authentication
        interceptors = []
        if self.auth is not None:
            creds_fn = self.auth.credentials
            interceptors.append(MetadataUnaryInterceptor(creds_fn))
            interceptors.append(MetadataUnaryStreamInterceptor(creds_fn))
            interceptors.append(MetadataStreamInterceptor(creds_fn))

        # Add the user agent to the options
        options = (("grpc.primary_user_agent", user_agent()),)

        if self.insecure:
            channel = aio.insecure_channel(
                self.addrport, interceptors=interceptors, options=options
            )
        else:
            channel = aio.secure_channel(
                self.addrport,
                credentials=grpc.ssl_channel_credentials(),
                interceptors=interceptors,
                options=options,
            )

        return channel


class Client:
    """
    Client defines the actual gRPC client that makes requests to an Ensign server.
    """

    def __init__(
        self,
        connection,
        client_id="",
        topic_cache=None,
        reconnect_tick=timedelta(milliseconds=750),
        reconnect_timeout=timedelta(minutes=5),
    ):
        """
        Create a new client for making requests to Ensign. The connection is not
        established until the connect() method is called.

        Parameters
        ----------
        connection : Connection
            The connection to the Ensign server
        """

        self.connection = connection
        self.channel = None
        self.publishers = {}
        self.subscribers = {}
        self.pool = None
        self.topics = topic_cache
        self.reconnect_tick = reconnect_tick
        self.reconnect_timeout = reconnect_timeout
        self.shutdown = None

        # Create client ID if not provided, which exists for the lifetime of the client
        # and persists across reconnects
        if client_id == "":
            self.client_id = str(ULID())
        else:
            self.client_id = client_id

    async def __aenter__(self):
        self._ensure_ready()

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    def _ensure_ready(self):
        """
        Create the gRPC channel and stub if not already connected. This is done at call
        time rather than when the client is created to ensure that user code runs in
        the same event loop as the gRPC library. This function also creates the worker
        pool and shutdown event signal if not already created.
        TODO: We could avoid this by requiring the Ensign client to be created in the
        event loop, e.g. async with Client(...) as client: ..., but the current syntax
        is a bit easier.
        """

        if not self.channel:
            self.channel = self.connection.create_channel()
            self.stub = ensign_pb2_grpc.EnsignStub(self.channel)

        if not self.pool:
            self.pool = WorkerPool(max_workers=10, max_queue_size=100)

        if not self.shutdown:
            self.shutdown = asyncio.Event()

    @catch_rpc_error
    async def publish(self, topic, events, on_ack=None, on_nack=None):
        # Ensure we have a gRPC channel
        self._ensure_ready()

        if self.shutdown.is_set():
            raise EnsignClientClosingError("client is closing")

        # Attempt to hash the topic, which fails if there is no topic ID
        try:
            topic_hash = hash(topic)
        except ValueError:
            topic_hash = None

        # Check if there is already an open publish stream for this topic
        if topic_hash and topic_hash in self.publishers:
            publisher = self.publishers[topic_hash]
        else:
            # Create the publish stream for this topic
            publisher = Publisher(
                self,
                topic,
                on_ack=on_ack,
                on_nack=on_nack,
                reconnect_tick=self.reconnect_tick,
                reconnect_timeout=self.reconnect_timeout,
            )

            # Connect to the publish stream
            # TODO: Distinguish between authentication errors, topic errors, and connection errors
            await publisher.connect()

            # After connect we should have the topic ID from the server, so compute the
            # hash and save the open publisher stream
            topic_hash = hash(topic)
            self.publishers[topic_hash] = publisher

            # Run the publisher as a concurrent task which handles reconnects
            await self.pool.schedule(
                publisher.run(),
                done_callback=lambda: self.publishers.pop(topic_hash, None),
            )

        # Create a concurrent task to queue the events from the user
        await self.pool.schedule(publisher.queue_events(events))

    @catch_rpc_error
    async def subscribe(self, topics, query="", consumer_group=None):
        # TODO: Support query parameters

        # Ensure we have a gRPC channel
        self._ensure_ready()

        if self.shutdown.is_set():
            raise EnsignClientClosingError("client is closing")

        # Create a hash of the topics
        topic_hash = frozenset([t for t in topics])

        # Check if there is already an open stream for these topics
        if topic_hash in self.subscribers:
            subscriber = self.subscribers[topic_hash]
        else:
            # Create the subscribe stream for these topics
            subscriber = Subscriber(
                self,
                topics,
                query=query_pb2.Query(query=query),
                consumer_group=consumer_group,
                reconnect_tick=self.reconnect_tick,
                reconnect_timeout=self.reconnect_timeout,
            )
            self.subscribers[topic_hash] = subscriber

            # Connect to the subscribe stream
            # TODO: Distinguish between authentication errors, topic errors, and connection errors
            await subscriber.connect()

            # Run the subscriber as a concurrent task which handles reconnects
            await self.pool.schedule(
                subscriber.run(),
                done_callback=lambda: self.subscribers.pop(topic_hash, None),
            )

        # Consume the events and yield them to the caller
        async for event in subscriber.consume():
            yield event

    @catch_rpc_error
    async def en_sql(self, query, params=None):
        self._ensure_ready()
        req = query_pb2.Query(query=query, params=params)
        return await Cursor.execute(self.stub.EnSQL(req))

    @catch_rpc_error
    async def explain(self, query="", params=None):
        self._ensure_ready()
        req = query_pb2.Query(query=query, params=params)
        return await self.stub.Explain(req)

    @catch_rpc_error
    async def list_topics(self, page_size=100, next_page_token=""):
        self._ensure_ready()
        params = ensign_pb2.PageInfo(
            page_size=page_size, next_page_token=next_page_token
        )
        rep = await self.stub.ListTopics(params)
        return rep.topics, rep.next_page_token

    @catch_rpc_error
    async def create_topic(self, topic):
        self._ensure_ready()
        return await self.stub.CreateTopic(topic)

    @catch_rpc_error
    async def retrieve_topic(self, id):
        self._ensure_ready()
        topic = topic_pb2.Topic(id=id)
        return await self.stub.RetrieveTopic(topic)

    @catch_rpc_error
    async def archive_topic(self, id):
        self._ensure_ready()
        params = topic_pb2.TopicMod(
            id=id, operation=topic_pb2.TopicMod.Operation.ARCHIVE
        )
        rep = await self.stub.DeleteTopic(params)
        return rep.id, rep.state

    @catch_rpc_error
    async def destroy_topic(self, id):
        self._ensure_ready()
        params = topic_pb2.TopicMod(
            id=id, operation=topic_pb2.TopicMod.Operation.DESTROY
        )
        rep = await self.stub.DeleteTopic(params)
        return rep.id, rep.state

    @catch_rpc_error
    async def topic_names(self, page_size=100, next_page_token=""):
        self._ensure_ready()
        params = ensign_pb2.PageInfo(
            page_size=page_size, next_page_token=next_page_token
        )
        rep = await self.stub.TopicNames(params)
        return rep.topic_names, rep.next_page_token

    @catch_rpc_error
    async def topic_exists(self, topic_id=None, project_id=None, topic_name=""):
        self._ensure_ready()
        params = topic_pb2.TopicName(
            topic_id=topic_id, project_id=project_id, name=topic_name
        )
        rep = await self.stub.TopicExists(params)
        return rep.query, rep.exists

    @catch_rpc_error
    async def info(self, topics=[]):
        self._ensure_ready()
        return await self.stub.Info(ensign_pb2.InfoRequest(topics=topics))

    @catch_rpc_error
    async def status(self, attempts=0, last_checked_at=None):
        self._ensure_ready()
        params = ensign_pb2.HealthCheck(
            attempts=attempts, last_checked_at=last_checked_at
        )
        rep = await self.stub.Status(params)
        return rep.status, rep.version, rep.uptime, rep.not_before, rep.not_after

    async def close(self):
        """
        Close the connection to the server and all ongoing streams.
        """
        if self.channel:
            await self.channel.close()
        await self._close_streams()
        self.channel = None

    async def _close_streams(self):
        # Prevent new streams from being created
        self.shutdown.set()

        # Close all ongoing publishers and subscribers
        while self.publishers:
            _, publisher = self.publishers.popitem()
            await publisher.close()

        while self.subscribers:
            _, subscriber = self.subscribers.popitem()
            await subscriber.close()

        await self.pool.release()


class Cursor:
    """
    Cursor defines an abstraction for server-side event streaming using familiar
    database cursor semantics.
    """

    @classmethod
    async def execute(cls, stream):
        self = Cursor()
        self._stream = stream
        self._result = None

        # Read the first event from the stream, which might raise an exception
        self._result = await self.fetchone()
        if self._result is None:
            raise QueryNoRows("no results returned by query")
        return self

    async def __aiter__(self):
        while True:
            event = await self.fetchone()
            if event is None:
                return
            yield event

    async def fetchone(self):
        """
        Fetch the next event from the stream. Returns None if there are no more events.
        """
        return await self._read()

    async def fetchmany(self, n):
        """
        Fetch the next n events from the stream. Returns an empty list if there are no
        more events.
        """
        events = []
        for _ in range(n):
            event = await self._read()
            if event is None:
                break
            events.append(event)
        return events

    async def fetchall(self):
        """
        Fetch all events from the stream. Returns an empty list if there are no more
        events.
        """
        events = []
        while True:
            event = await self._read()
            if event is None:
                break
            events.append(event)
        return events

    def close(self):
        """
        Close the cursor and the underlying stream.
        """
        if self._stream is not None:
            self._stream.cancel()
            self._stream = None

    async def _read(self):
        """
        Read the next event on the stream, handling gRPC errors and the end of the
        stream.
        """
        if self._stream is None:
            return None

        # Return the cached result if there is one
        if self._result is not None:
            result = self._result
            self._result = None
            return result

        # Read the next event from the stream
        try:
            event = await self._stream.read()
        except grpc.aio.AioRpcError as e:
            self.close()
            if e.code() == grpc.StatusCode.CANCELLED:
                return None
            else:
                raise e
        except (asyncio.CancelledError, StopAsyncIteration):
            self.close()
            return None

        # Handle unexpected end of stream
        if event is grpc.aio.EOF:
            self.close()
            return None

        # Convert the event to the user-facing type
        return from_proto(unwrap(event))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: region/v1beta1/region.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bregion/v1beta1/region.proto\x12\x0eregion.v1beta1\x1a\x1fgoogle/protobuf/timestamp.proto"\xd8\x01\n\nRegionInfo\x12"\n\x02id\x18\x01 \x01(\x0e\x32\x16.region.v1beta1.Region\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07\x63ountry\x18\x03 \x01(\t\x12\x0c\n\x04zone\x18\x04 \x01(\t\x12\r\n\x05\x63loud\x18\x05 \x01(\t\x12\x0f\n\x07\x63luster\x18\x06 \x01(\t\x12,\n\x08modified\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12+\n\x07\x63reated\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp*\xdb\x1a\n\x06Region\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x16\n\x12STG_LKE_US_EAST_1A\x10\x01\x12\x14\n\x0eLKE_AP_WEST_1A\x10\xe0\xe1R\x12\x17\n\x11LKE_CA_CENTRAL_1A\x10\xaa\xcd\x44\x12\x19\n\x13LKE_AP_SOUTHEAST_1A\x10\xf4\x9d?\x12\x17\n\x11LKE_US_CENTRAL_1A\x10\x9e\xa7p\x12\x14\n\x0eLKE_US_WEST_1A\x10\xa8\xa7p\x12\x19\n\x13LKE_US_SOUTHEAST_1A\x10\xb2\xa7p\x12\x14\n\x0eLKE_US_EAST_1A\x10\xbc\xa7p\x12\x14\n\x0eLKE_EU_WEST_1A\x10\x96\xbao\x12\x15\n\x0fLKE_AP_SOUTH_1A\x10\xc0\xf1g\x12\x17\n\x11LKE_EU_CENTRAL_1A\x10\xba\xf1M\x12\x19\n\x13LKE_AP_NORTHEAST_1A\x10\xe4\xfbT\x12\x17\n\x10GCP_ASIA_EAST_1A\x10\xb1\xdb\x83\x01\x12\x17\n\x10GCP_ASIA_EAST_1B\x10\xb2\xdb\x83\x01\x12\x17\n\x10GCP_ASIA_EAST_1C\x10\xb3\xdb\x83\x01\x12\x17\n\x10GCP_ASIA_EAST_2A\x10\xca\x88\x8f\x01\x12\x17\n\x10GCP_ASIA_EAST_2B\x10\xcb\x88\x8f\x01\x12\x17\n\x10GCP_ASIA_EAST_2C\x10\xcc\x88\x8f\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_1A\x10\xd4\xff\x91\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_1B\x10\xd5\xff\x91\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_1C\x10\xd6\xff\x91\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_2A\x10\xde\xff\x91\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_2B\x10\xdf\xff\x91\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_2C\x10\xe0\xff\x91\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_3A\x10\xb8\x8c\x93\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_3B\x10\xb9\x8c\x93\x01\x12\x1c\n\x15GCP_ASIA_NORTHEAST_3C\x10\xba\x8c\x93\x01\x12\x18\n\x11GCP_ASIA_SOUTH_1A\x10\xd3\xe6\x8f\x01\x12\x18\n\x11GCP_ASIA_SOUTH_1B\x10\xd4\xe6\x8f\x01\x12\x18\n\x11GCP_ASIA_SOUTH_1C\x10\xd5\xe6\x8f\x01\x12\x18\n\x11GCP_ASIA_SOUTH_2A\x10\xdd\xe6\x8f\x01\x12\x18\n\x11GCP_ASIA_SOUTH_2B\x10\xde\xe6\x8f\x01\x12\x18\n\x11GCP_ASIA_SOUTH_2C\x10\xdf\xe6\x8f\x01\x12\x1c\n\x15GCP_ASIA_SOUTHEAST_1A\x10\xf6\xf5\xa4\x01\x12\x1c\n\x15GCP_ASIA_SOUTHEAST_1B\x10\xf7\xf5\xa4\x01\x12\x1c\n\x15GCP_ASIA_SOUTHEAST_1C\x10\xf8\xf5\xa4\x01\x12\x1c\n\x15GCP_ASIA_SOUTHEAST_2A\x10\x90\x86\x90\x01\x12\x1c\n\x15GCP_ASIA_SOUTHEAST_2B\x10\x91\x86\x90\x01\x12\x1c\n\x15GCP_ASIA_SOUTHEAST_2C\x10\x92\x86\x90\x01\x12 \n\x1aGCP_AUSTRALIA_SOUTHEAST_1A\x10\xfa\xa2|\x12 \n\x1aGCP_AUSTRALIA_SOUTHEAST_1B\x10\xfb\xa2|\x12 \n\x1aGCP_AUSTRALIA_SOUTHEAST_1C\x10\xfc\xa2|\x12 \n\x1aGCP_AUSTRALIA_SOUTHEAST_2A\x10\x84\xa3|\x12 \n\x1aGCP_AUSTRALIA_SOUTHEAST_2B\x10\x85\xa3|\x12 \n\x1aGCP_AUSTRALIA_SOUTHEAST_2C\x10\x86\xa3|\x12\x1c\n\x15GCP_EUROPE_CENTRAL_2A\x10\xae\xd6\x9f\x01\x12\x1c\n\x15GCP_EUROPE_CENTRAL_2B\x10\xaf\xd6\x9f\x01\x12\x1c\n\x15GCP_EUROPE_CENTRAL_2C\x10\xb0\xd6\x9f\x01\x12\x1a\n\x13GCP_EUROPE_NORTH_1A\x10\xe8\x8b\x89\x01\x12\x1a\n\x13GCP_EUROPE_NORTH_1B\x10\xe9\x8b\x89\x01\x12\x1a\n\x13GCP_EUROPE_NORTH_1C\x10\xea\x8b\x89\x01\x12\x1e\n\x17GCP_EUROPE_SOUTHWEST_1A\x10\xa2\xa2\xa6\x01\x12\x1e\n\x17GCP_EUROPE_SOUTHWEST_1B\x10\xa3\xa2\xa6\x01\x12\x1e\n\x17GCP_EUROPE_SOUTHWEST_1C\x10\xa4\xa2\xa6\x01\x12\x18\n\x12GCP_EUROPE_WEST_1B\x10\xcd\xbf}\x12\x18\n\x12GCP_EUROPE_WEST_1C\x10\xce\xbf}\x12\x18\n\x12GCP_EUROPE_WEST_1D\x10\xcf\xbf}\x12\x1a\n\x13GCP_EUROPE_WEST_12A\x10\xf6\xa2\x91\x01\x12\x1a\n\x13GCP_EUROPE_WEST_12B\x10\xf7\xa2\x91\x01\x12\x1a\n\x13GCP_EUROPE_WEST_12C\x10\xf8\xa2\x91\x01\x12\x19\n\x12GCP_EUROPE_WEST_2A\x10\xb0\xbf\xac\x01\x12\x19\n\x12GCP_EUROPE_WEST_2B\x10\xb1\xbf\xac\x01\x12\x19\n\x12GCP_EUROPE_WEST_2C\x10\xb2\xbf\xac\x01\x12\x19\n\x12GCP_EUROPE_WEST_3A\x10\xca\xf6\x8a\x01\x12\x19\n\x12GCP_EUROPE_WEST_3B\x10\xcb\xf6\x8a\x01\x12\x19\n\x12GCP_EUROPE_WEST_3C\x10\xcc\xf6\x8a\x01\x12\x19\n\x12GCP_EUROPE_WEST_4A\x10\xb4\xa7\x9a\x01\x12\x19\n\x12GCP_EUROPE_WEST_4B\x10\xb5\xa7\x9a\x01\x12\x19\n\x12GCP_EUROPE_WEST_4C\x10\xb6\xa7\x9a\x01\x12\x19\n\x12GCP_EUROPE_WEST_6A\x10\xde\x9c\xa8\x01\x12\x19\n\x12GCP_EUROPE_WEST_6B\x10\xdf\x9c\xa8\x01\x12\x19\n\x12GCP_EUROPE_WEST_6C\x10\xe0\x9c\xa8\x01\x12\x19\n\x12GCP_EUROPE_WEST_8A\x10\xa8\xa3\x91\x01\x12\x19\n\x12GCP_EUROPE_WEST_8B\x10\xa9\xa3\x91\x01\x12\x19\n\x12GCP_EUROPE_WEST_8C\x10\xaa\xa3\x91\x01\x12\x19\n\x12GCP_EUROPE_WEST_9A\x10\xe2\xab\x89\x01\x12\x19\n\x12GCP_EUROPE_WEST_9B\x10\xe3\xab\x89\x01\x12\x19\n\x12GCP_EUROPE_WEST_9C\x10\xe4\xab\x89\x01\x12\x18\n\x11GCP_ME_CENTRAL_1A\x10\xbc\xa6\x90\x01\x12\x18\n\x11GCP_ME_CENTRAL_1B\x10\xbd\xa6\x90\x01\x12\x18\n\x11GCP_ME_CENTRAL_1C\x10\xbe\xa6\x90\x01\x12\x15\n\x0eGCP_ME_WEST_1A\x10\xa6\x84\x91\x01\x12\x15\n\x0eGCP_ME_WEST_1B\x10\xa7\x84\x91\x01\x12\x15\n\x0eGCP_ME_WEST_1C\x10\xa8\x84\x91\x01\x12$\n\x1dGCP_NORTHAMERICA_NORTHEAST_1A\x10\xd0\xd3\x81\x01\x12$\n\x1dGCP_NORTHAMERICA_NORTHEAST_1B\x10\xd1\xd3\x81\x01\x12$\n\x1dGCP_NORTHAMERICA_NORTHEAST_1C\x10\xd2\xd3\x81\x01\x12$\n\x1dGCP_NORTHAMERICA_NORTHEAST_2A\x10\xda\xd3\x81\x01\x12$\n\x1dGCP_NORTHAMERICA_NORTHEAST_2B\x10\xdb\xd3\x81\x01\x12$\n\x1dGCP_NORTHAMERICA_NORTHEAST_2C\x10\xdc\xd3\x81\x01\x12\x1e\n\x18GCP_SOUTHAMERICA_EAST_1A\x10\xe4\xdc~\x12\x1e\n\x18GCP_SOUTHAMERICA_EAST_1B\x10\xe5\xdc~\x12\x1e\n\x18GCP_SOUTHAMERICA_EAST_1C\x10\xe6\xdc~\x12\x1f\n\x18GCP_SOUTHAMERICA_WEST_1A\x10\xce\xae\x83\x01\x12\x1f\n\x18GCP_SOUTHAMERICA_WEST_1B\x10\xcf\xae\x83\x01\x12\x1f\n\x18GCP_SOUTHAMERICA_WEST_1C\x10\xd0\xae\x83\x01\x12\x18\n\x11GCP_US_CENTRAL_1A\x10\xd8\xad\xad\x01\x12\x18\n\x11GCP_US_CENTRAL_1B\x10\xd9\xad\xad\x01\x12\x18\n\x11GCP_US_CENTRAL_1C\x10\xda\xad\xad\x01\x12\x18\n\x11GCP_US_CENTRAL_1F\x10\xdd\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_1B\x10\xe3\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_1C\x10\xe4\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_1D\x10\xe5\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_4A\x10\xec\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_4B\x10\xed\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_4C\x10\xee\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_5A\x10\xf6\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_5B\x10\xf7\xad\xad\x01\x12\x15\n\x0eGCP_US_EAST_5C\x10\xf8\xad\xad\x01\x12\x16\n\x0fGCP_US_SOUTH_1A\x10\x80\xae\xad\x01\x12\x16\n\x0fGCP_US_SOUTH_1B\x10\x81\xae\xad\x01\x12\x16\n\x0fGCP_US_SOUTH_1C\x10\x82\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_1A\x10\x8a\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_1B\x10\x8b\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_1C\x10\x8c\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_2A\x10\x94\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_2B\x10\x95\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_2C\x10\x96\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_3A\x10\x9e\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_3B\x10\x9f\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_3C\x10\xa0\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_4A\x10\xa8\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_4B\x10\xa9\xae\xad\x01\x12\x15\n\x0eGCP_US_WEST_4C\x10\xaa\xae\xad\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "region.v1beta1.region_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _REGION._serialized_start = 300
    _REGION._serialized_end = 3719
    _REGIONINFO._serialized_start = 81
    _REGIONINFO._serialized_end = 297
# @@protoc_insertion_point(module_scope)



import grpc
import asyncio
from datetime import datetime, timedelta

from ulid import ULID
from pyensign.events import from_proto
from pyensign.api.v1beta1 import ensign_pb2
from pyensign.utils.queue import BidiQueue
from pyensign.api.v1beta1.event import wrap, unwrap
from pyensign.exceptions import (
    EnsignError,
    EnsignTypeError,
    EnsignTimeoutError,
    EnsignInitError,
    EnsignInvalidTopicError,
    EnsignTopicNotFoundError,
)
from pyensign.iterator import (
    RequestIterator,
    PublishResponseIterator,
    SubscribeResponseIterator,
)

RECONNECT_TICK = timedelta(milliseconds=750)
RECONNECT_TIMEOUT = timedelta(minutes=5)


class StreamHandler:
    """
    StreamHandler manages the connection to a gRPC stream and handles reconnections if
    the stream goes down.
    """

    def __init__(
        self,
        client,
        queue,
        reconnect_tick=RECONNECT_TICK,
        reconnect_timeout=RECONNECT_TIMEOUT,
    ):
        self.client_id = client.client_id
        self.stub = client.stub
        self.topic_cache = client.topics
        self.queue = queue
        self.reconnect_tick = reconnect_tick
        self.reconnect_timeout = reconnect_timeout
        self.shutdown = asyncio.Event()
        self._responses = None
        self._topics = {}

    async def connect(self):
        """
        Implement this method for protocol-specific stream initialization.
        """

        raise NotImplementedError

    async def run(self):
        """
        Send and recv on the stream and attempt to reconnect to the stream on failure.
        """

        if self._responses is None:
            raise EnsignInitError("stream is not connected")

        while not self.shutdown.is_set():
            # Consume the responses until a gRPC error occurs
            await self._responses.consume()

            # Close the queue to stop the current request iterator
            await self.queue.close()

            # Try to reconnect to the stream
            try:
                await self.reconnect()
            except EnsignError as e:
                # Return timeout errors etc. to the user by writing them to the queue
                await self.queue.write_response(e)
                return

    async def reconnect(self):
        """
        Try to reconnect to the stream until the timeout expires.
        """

        self.reset_timeout()
        while not self.timeout_expired():
            # Wait for the reconnect tick
            await asyncio.sleep(self.reconnect_tick.microseconds / 1000000)

            # Don't try to reconnect if the stream is being closed
            if self.shutdown.is_set():
                return

            # Try to connect and ignore gRPC errors
            try:
                await self.connect()
                return
            except grpc.aio.AioRpcError:
                continue
            except asyncio.CancelledError:
                # Respect task cancellation from gRPC, which usually means that the
                # underlying channel was closed by the client. At this point it won't
                # be possible to reconnect to the stream.
                await self.close()
                return

        # Timeout expired, give up.
        raise EnsignTimeoutError("timeout expired while trying to reconnect to stream")

    def _parse_topics(self, topic_map):
        """
        Parse topics from a topic map, updating the state on the stream and the global
        topic cache.
        """

        for name, id_bytes in topic_map.items():
            try:
                id = ULID(id_bytes)
            except ValueError:
                # Ignore unparseable topic IDs
                continue

            # Update the topic cache
            self._topics[name] = id
            self.topic_cache.add(name, id)

    def reset_timeout(self):
        self.timeout = datetime.now() + self.reconnect_timeout

    def timeout_expired(self):
        return datetime.now() > self.timeout

    async def close(self):
        """
        Close the stream as gracefully as possible.
        """

        self.shutdown.set()
        await self.queue.close()


class Publisher(StreamHandler):
    """
    Publisher manages the connection to a gRPC publish stream. Once the ready message
    is received from the server, the stream is considered open and the request/response
    iterators are used to send and receive messages on the stream. The publisher relies
    on the response iterator to detect when the stream has been closed. If the iterator
    is exhausted, the publisher attempts to reconnect to the stream. If the reconnect
    timeout is exceeded then the publisher gives up and closes the stream.
    """

    def __init__(
        self,
        client,
        topic,
        on_ack=None,
        on_nack=None,
        **kwargs,
    ):
        super().__init__(client, BidiQueue(), **kwargs)
        self.topic = topic
        self.on_ack = on_ack
        self.on_nack = on_nack
        self.pending = {}

    async def connect(self):
        """
        Attempt to establish a connection to the server and raise an exception if the
        connection could not be established.
        """

        # Create the gRPC stream from the request iterator
        open_stream = ensign_pb2.PublisherRequest(
            open_stream=ensign_pb2.OpenStream(client_id=self.client_id)
        )
        stream = self.stub.Publish(RequestIterator(self.queue, open_stream))

        # First response from the server should be a ready message
        rep = await stream.read()
        rep_type = rep.WhichOneof("embed")
        if rep_type != "ready":
            raise EnsignTypeError("expected ready response, got {}".format(rep_type))

        # Save the topics from the ready message
        self._parse_topics(rep.ready.topics)

        # Set the topic ID for publish requests
        try:
            self._update_topic()
        except EnsignTopicNotFoundError as e:
            await self.close()
            raise e

        # Create the response iterator from the gRPC stream
        self._responses = PublishResponseIterator(
            stream, self.pending, on_ack=self.on_ack, on_nack=self.on_nack
        )

    async def queue_events(self, events):
        """
        Queue events to be published on the stream.
        """

        async for event in events:
            # If we're shutting down, stop queueing events
            if self.shutdown.is_set():
                break

            # Queue the event
            wrapper = wrap(event.proto(), self.topic.id)
            req = ensign_pb2.PublisherRequest(event=wrapper)
            await self.queue.write_request(req)
            event.mark_published()

            # Save the event for ack/nack handling
            # TODO: How do we handle events that are never acked/nacked?
            self.pending[wrapper.local_id] = event

    def _update_topic(self):
        """
        Set the topic ID for the publisher by performing a lookup into the map. An
        exception is raised if the topic is not in the map, which implies that the
        topic does not exist or it was created after the stream was opened.
        """

        if self.topic.id:
            # If topic ID is already set, ensure that it exists
            # Edge case: The topic name looks like a topic ID
            if (
                self.topic.id not in self._topics.values()
                and str(self.topic.id) not in self._topics
            ):
                raise EnsignTopicNotFoundError(self.topic.id)
        elif self.topic.name:
            self.topic.id = self._topics.get(self.topic.name, None)
            if not self.topic.id:
                raise EnsignTopicNotFoundError(self.topic.id)
        else:
            raise EnsignInvalidTopicError("topic has no name or ID")


class Subscriber(StreamHandler):
    """
    Subscriber manages the connection to a gRPC stream. Once the ready message is
    received from the server, the stream is considered open and the request/response
    iterators are used to send and receive messages on the stream. The subscriber
    relies on the response iterator to detect when the stream has been closed. If the
    iterator is exhausted, the subscriber attempts to reconnect to the stream. If
    the reconnect timeout is exceeded then the subscriber gives up and closes the
    stream.
    """

    def __init__(
        self,
        client,
        topics,
        query="",
        consumer_group=None,
        **kwargs,
    ):
        super().__init__(client, BidiQueue(), **kwargs)
        self.topics = topics
        self.query = query
        self.consumer_group = consumer_group

    async def connect(self):
        """
        Attempt to establish a connection to the server and raise an exception if the
        connection could not be established.
        """

        # Create the gRPC stream from the request iterator
        sub = ensign_pb2.SubscribeRequest(
            subscription=ensign_pb2.Subscription(
                client_id=self.client_id,
                topics=self.topics,
                query=self.query,
                group=self.consumer_group,
            )
        )
        stream = self.stub.Subscribe(RequestIterator(self.queue, sub))

        # First response from the server should be a ready message
        rep = await stream.read()
        rep_type = rep.WhichOneof("embed")
        if rep_type != "ready":
            raise EnsignTypeError("expected ready response, got {}".format(rep_type))

        # Save the topics from the ready message
        self._parse_topics(rep.ready.topics)

        # Create the response iterator from the gRPC stream
        self._responses = SubscribeResponseIterator(stream, self.queue)

    async def consume(self):
        """
        Consume the events from the incoming queue and yield them to the caller. This
        coroutine is completely independent from the gRPC stream, so it will continue
        to run even if the stream is closed. This allows the subscriber to reconnect to
        the stream in the background without interrupting any event processing. The
        caller must handle EnsignError exceptions.
        """

        while True:
            rep = await self.queue.read_response()
            if rep is None:
                if self.shutdown.is_set():
                    # If the subscriber is closing, stop consuming events
                    break
                else:
                    # Otherwise, wait for reconnect
                    continue
            elif isinstance(rep, EnsignError):
                # Raise errors to the caller, this includes connection timeout errors
                # and protocol errors
                raise rep
            else:
                # Convert the event into the user facing type
                event = from_proto(unwrap(rep))
                event.mark_subscribed(rep.id, self.queue)
                yield event



from ulid import ULID

from pyensign.api.v1beta1 import event_pb2


def wrap(event, topic_id):
    """
    Wrap an event and topic ULID into an EventWrapper.
    """

    if not isinstance(event, event_pb2.Event):
        raise TypeError("event must be a protobuf event")

    if not isinstance(topic_id, ULID):
        raise TypeError("topic_id must be a ULID")

    return event_pb2.EventWrapper(
        local_id=ULID().bytes, event=event.SerializeToString(), topic_id=topic_id.bytes
    )


def unwrap(event_wrapper):
    """
    Unwrap an event from an EventWrapper.
    """

    if event_wrapper.event is None or len(event_wrapper.event) == 0:
        raise AttributeError("event wrapper contains no event")

    event = event_pb2.Event()
    event.ParseFromString(event_wrapper.event)
    return event



from grpc import aio, StatusCode

from pyensign.exceptions import AuthenticationError


class MetadataInterceptor:
    """
    MetadataInterceptor is the base class for interceptors that add metadata to each
    gRPC request (e.g. authentication credentials)
    """

    def __init__(self, metadata_fn):
        self.metadata_fn = metadata_fn

    def get_metadata(self, client_call_details):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        try:
            metadata.append(self.metadata_fn())
        except AuthenticationError as e:
            # TODO: gRPC wants either an AioRpcError or an asyncio.CancelledError,
            # this is a bit of a hack to get gRPC to handle the exception, so users
            # will see an EnsignRPCError instead of the AuthenticationError
            raise aio.AioRpcError(
                StatusCode.UNAUTHENTICATED,
                initial_metadata=None,
                trailing_metadata=None,
                details=str(e),
            )
        return metadata


class MetadataUnaryInterceptor(MetadataInterceptor, aio.UnaryUnaryClientInterceptor):
    async def intercept_unary_unary(self, continuation, client_call_details, request):
        """
        Intercepts a unary-unary call to add a metadata tuple.
        """

        client_call_details = client_call_details._replace(
            metadata=self.get_metadata(client_call_details)
        )
        return await continuation(client_call_details, request)


class MetadataUnaryStreamInterceptor(
    MetadataInterceptor, aio.UnaryStreamClientInterceptor
):
    async def intercept_unary_stream(self, continuation, client_call_details, request):
        """
        Intercepts a unary-stream call to add a metadata tuple.
        """

        client_call_details = client_call_details._replace(
            metadata=self.get_metadata(client_call_details)
        )
        return await continuation(client_call_details, request)


class MetadataStreamInterceptor(MetadataInterceptor, aio.StreamStreamClientInterceptor):
    async def intercept_stream_stream(
        self, continuation, client_call_details, request_iterator
    ):
        """
        Intercepts a stream-stream call to add a metadata tuple.
        """

        client_call_details = client_call_details._replace(
            metadata=self.get_metadata(client_call_details)
        )
        return await continuation(client_call_details, request_iterator)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from pyensign.api.v1beta1 import ensign_pb2 as api_dot_v1beta1_dot_ensign__pb2
from pyensign.api.v1beta1 import event_pb2 as api_dot_v1beta1_dot_event__pb2
from pyensign.api.v1beta1 import query_pb2 as api_dot_v1beta1_dot_query__pb2
from pyensign.api.v1beta1 import topic_pb2 as api_dot_v1beta1_dot_topic__pb2


class EnsignStub(object):
    """The Ensign service is meant to allow publishers (producers) and subscribers
    (consumers) of events to interact with the Ensign eventing system; e.g. this is a
    user-oriented API that is the basis of the user SDKs that we will build. There are
    two primary interactions that the user client may have: publishing or subscribing to
    topics to send and receive events or managing topics that are available.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Publish = channel.stream_stream(
            "/ensign.v1beta1.Ensign/Publish",
            request_serializer=api_dot_v1beta1_dot_ensign__pb2.PublisherRequest.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_ensign__pb2.PublisherReply.FromString,
        )
        self.Subscribe = channel.stream_stream(
            "/ensign.v1beta1.Ensign/Subscribe",
            request_serializer=api_dot_v1beta1_dot_ensign__pb2.SubscribeRequest.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_ensign__pb2.SubscribeReply.FromString,
        )
        self.EnSQL = channel.unary_stream(
            "/ensign.v1beta1.Ensign/EnSQL",
            request_serializer=api_dot_v1beta1_dot_query__pb2.Query.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_event__pb2.EventWrapper.FromString,
        )
        self.Explain = channel.unary_unary(
            "/ensign.v1beta1.Ensign/Explain",
            request_serializer=api_dot_v1beta1_dot_query__pb2.Query.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_query__pb2.QueryExplanation.FromString,
        )
        self.ListTopics = channel.unary_unary(
            "/ensign.v1beta1.Ensign/ListTopics",
            request_serializer=api_dot_v1beta1_dot_ensign__pb2.PageInfo.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_topic__pb2.TopicsPage.FromString,
        )
        self.CreateTopic = channel.unary_unary(
            "/ensign.v1beta1.Ensign/CreateTopic",
            request_serializer=api_dot_v1beta1_dot_topic__pb2.Topic.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_topic__pb2.Topic.FromString,
        )
        self.RetrieveTopic = channel.unary_unary(
            "/ensign.v1beta1.Ensign/RetrieveTopic",
            request_serializer=api_dot_v1beta1_dot_topic__pb2.Topic.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_topic__pb2.Topic.FromString,
        )
        self.DeleteTopic = channel.unary_unary(
            "/ensign.v1beta1.Ensign/DeleteTopic",
            request_serializer=api_dot_v1beta1_dot_topic__pb2.TopicMod.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_topic__pb2.TopicTombstone.FromString,
        )
        self.TopicNames = channel.unary_unary(
            "/ensign.v1beta1.Ensign/TopicNames",
            request_serializer=api_dot_v1beta1_dot_ensign__pb2.PageInfo.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_topic__pb2.TopicNamesPage.FromString,
        )
        self.TopicExists = channel.unary_unary(
            "/ensign.v1beta1.Ensign/TopicExists",
            request_serializer=api_dot_v1beta1_dot_topic__pb2.TopicName.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_topic__pb2.TopicExistsInfo.FromString,
        )
        self.Info = channel.unary_unary(
            "/ensign.v1beta1.Ensign/Info",
            request_serializer=api_dot_v1beta1_dot_ensign__pb2.InfoRequest.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_ensign__pb2.ProjectInfo.FromString,
        )
        self.Status = channel.unary_unary(
            "/ensign.v1beta1.Ensign/Status",
            request_serializer=api_dot_v1beta1_dot_ensign__pb2.HealthCheck.SerializeToString,
            response_deserializer=api_dot_v1beta1_dot_ensign__pb2.ServiceState.FromString,
        )


class EnsignServicer(object):
    """The Ensign service is meant to allow publishers (producers) and subscribers
    (consumers) of events to interact with the Ensign eventing system; e.g. this is a
    user-oriented API that is the basis of the user SDKs that we will build. There are
    two primary interactions that the user client may have: publishing or subscribing to
    topics to send and receive events or managing topics that are available.
    """

    def Publish(self, request_iterator, context):
        """Both the Publish and Subscribe RPCs are bidirectional streaming to allow for acks
        and nacks of events to be sent between Ensign and the client. The Publish stream
        is opened and the client sends events and receives acks/nacks -- when the client
        closes the publish stream, the server sends back information about the current
        state of the topic. When the Subscribe stream is opened, the client must send an
        open stream message with the subscription info before receiving events. Once it
        receives events it must send back acks/nacks up the stream so that Ensign
        advances the topic offset for the rest of the clients in the group.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Subscribe(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def EnSQL(self, request, context):
        """EnSQL is a server-side streaming RPC that executes an query and returns a stream
        of events as a result set back from the query. It terminates once all results
        have been returned or the client terminates the stream.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Explain(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListTopics(self, request, context):
        """This is a simple topic management interface. Right now we assume that topics are
        immutable, therefore there is no update topic RPC call. There are two ways to
        delete a topic - archiving it makes the topic readonly so that no events can be
        published to it, but it can still be read. Destroying the topic deletes it and
        removes all of its data, freeing up the topic name to be used again.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateTopic(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RetrieveTopic(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def DeleteTopic(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def TopicNames(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def TopicExists(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Info(self, request, context):
        """Info provides statistics and metrics describing the state of a project"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Status(self, request, context):
        """Implements a client-side heartbeat that can also be used by monitoring tools."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_EnsignServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Publish": grpc.stream_stream_rpc_method_handler(
            servicer.Publish,
            request_deserializer=api_dot_v1beta1_dot_ensign__pb2.PublisherRequest.FromString,
            response_serializer=api_dot_v1beta1_dot_ensign__pb2.PublisherReply.SerializeToString,
        ),
        "Subscribe": grpc.stream_stream_rpc_method_handler(
            servicer.Subscribe,
            request_deserializer=api_dot_v1beta1_dot_ensign__pb2.SubscribeRequest.FromString,
            response_serializer=api_dot_v1beta1_dot_ensign__pb2.SubscribeReply.SerializeToString,
        ),
        "EnSQL": grpc.unary_stream_rpc_method_handler(
            servicer.EnSQL,
            request_deserializer=api_dot_v1beta1_dot_query__pb2.Query.FromString,
            response_serializer=api_dot_v1beta1_dot_event__pb2.EventWrapper.SerializeToString,
        ),
        "Explain": grpc.unary_unary_rpc_method_handler(
            servicer.Explain,
            request_deserializer=api_dot_v1beta1_dot_query__pb2.Query.FromString,
            response_serializer=api_dot_v1beta1_dot_query__pb2.QueryExplanation.SerializeToString,
        ),
        "ListTopics": grpc.unary_unary_rpc_method_handler(
            servicer.ListTopics,
            request_deserializer=api_dot_v1beta1_dot_ensign__pb2.PageInfo.FromString,
            response_serializer=api_dot_v1beta1_dot_topic__pb2.TopicsPage.SerializeToString,
        ),
        "CreateTopic": grpc.unary_unary_rpc_method_handler(
            servicer.CreateTopic,
            request_deserializer=api_dot_v1beta1_dot_topic__pb2.Topic.FromString,
            response_serializer=api_dot_v1beta1_dot_topic__pb2.Topic.SerializeToString,
        ),
        "RetrieveTopic": grpc.unary_unary_rpc_method_handler(
            servicer.RetrieveTopic,
            request_deserializer=api_dot_v1beta1_dot_topic__pb2.Topic.FromString,
            response_serializer=api_dot_v1beta1_dot_topic__pb2.Topic.SerializeToString,
        ),
        "DeleteTopic": grpc.unary_unary_rpc_method_handler(
            servicer.DeleteTopic,
            request_deserializer=api_dot_v1beta1_dot_topic__pb2.TopicMod.FromString,
            response_serializer=api_dot_v1beta1_dot_topic__pb2.TopicTombstone.SerializeToString,
        ),
        "TopicNames": grpc.unary_unary_rpc_method_handler(
            servicer.TopicNames,
            request_deserializer=api_dot_v1beta1_dot_ensign__pb2.PageInfo.FromString,
            response_serializer=api_dot_v1beta1_dot_topic__pb2.TopicNamesPage.SerializeToString,
        ),
        "TopicExists": grpc.unary_unary_rpc_method_handler(
            servicer.TopicExists,
            request_deserializer=api_dot_v1beta1_dot_topic__pb2.TopicName.FromString,
            response_serializer=api_dot_v1beta1_dot_topic__pb2.TopicExistsInfo.SerializeToString,
        ),
        "Info": grpc.unary_unary_rpc_method_handler(
            servicer.Info,
            request_deserializer=api_dot_v1beta1_dot_ensign__pb2.InfoRequest.FromString,
            response_serializer=api_dot_v1beta1_dot_ensign__pb2.ProjectInfo.SerializeToString,
        ),
        "Status": grpc.unary_unary_rpc_method_handler(
            servicer.Status,
            request_deserializer=api_dot_v1beta1_dot_ensign__pb2.HealthCheck.FromString,
            response_serializer=api_dot_v1beta1_dot_ensign__pb2.ServiceState.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "ensign.v1beta1.Ensign", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Ensign(object):
    """The Ensign service is meant to allow publishers (producers) and subscribers
    (consumers) of events to interact with the Ensign eventing system; e.g. this is a
    user-oriented API that is the basis of the user SDKs that we will build. There are
    two primary interactions that the user client may have: publishing or subscribing to
    topics to send and receive events or managing topics that are available.
    """

    @staticmethod
    def Publish(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            "/ensign.v1beta1.Ensign/Publish",
            api_dot_v1beta1_dot_ensign__pb2.PublisherRequest.SerializeToString,
            api_dot_v1beta1_dot_ensign__pb2.PublisherReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Subscribe(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            "/ensign.v1beta1.Ensign/Subscribe",
            api_dot_v1beta1_dot_ensign__pb2.SubscribeRequest.SerializeToString,
            api_dot_v1beta1_dot_ensign__pb2.SubscribeReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def EnSQL(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/ensign.v1beta1.Ensign/EnSQL",
            api_dot_v1beta1_dot_query__pb2.Query.SerializeToString,
            api_dot_v1beta1_dot_event__pb2.EventWrapper.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Explain(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/Explain",
            api_dot_v1beta1_dot_query__pb2.Query.SerializeToString,
            api_dot_v1beta1_dot_query__pb2.QueryExplanation.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListTopics(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/ListTopics",
            api_dot_v1beta1_dot_ensign__pb2.PageInfo.SerializeToString,
            api_dot_v1beta1_dot_topic__pb2.TopicsPage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateTopic(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/CreateTopic",
            api_dot_v1beta1_dot_topic__pb2.Topic.SerializeToString,
            api_dot_v1beta1_dot_topic__pb2.Topic.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RetrieveTopic(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/RetrieveTopic",
            api_dot_v1beta1_dot_topic__pb2.Topic.SerializeToString,
            api_dot_v1beta1_dot_topic__pb2.Topic.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def DeleteTopic(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/DeleteTopic",
            api_dot_v1beta1_dot_topic__pb2.TopicMod.SerializeToString,
            api_dot_v1beta1_dot_topic__pb2.TopicTombstone.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def TopicNames(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/TopicNames",
            api_dot_v1beta1_dot_ensign__pb2.PageInfo.SerializeToString,
            api_dot_v1beta1_dot_topic__pb2.TopicNamesPage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def TopicExists(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/TopicExists",
            api_dot_v1beta1_dot_topic__pb2.TopicName.SerializeToString,
            api_dot_v1beta1_dot_topic__pb2.TopicExistsInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Info(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/Info",
            api_dot_v1beta1_dot_ensign__pb2.InfoRequest.SerializeToString,
            api_dot_v1beta1_dot_ensign__pb2.ProjectInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Status(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/ensign.v1beta1.Ensign/Status",
            api_dot_v1beta1_dot_ensign__pb2.HealthCheck.SerializeToString,
            api_dot_v1beta1_dot_ensign__pb2.ServiceState.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from pyensign.api.v1beta1.ensign_pb2 import Nack

Unknown = Nack.Code.UNKNOWN

# Client-side error codes. Subscribers should use these codes when nack-ing events to
# ensure compatibility with Ensign.
Unprocessed = Nack.Code.UNPROCESSED
Timeout = Nack.Code.TIMEOUT
UnhandledMimetype = Nack.Code.UNHANDLED_MIMETYPE
UnknownType = Nack.Code.UNKNOWN_TYPE
DeliverAgainAny = Nack.Code.DELIVER_AGAIN_ANY
DeliverAgainNotMe = Nack.Code.DELIVER_AGAIN_NOT_ME

# Server-side error codes. These codes are returned by the Ensign service. Publishers
# should inspect the error code in a Nack returned by the Ensign service to determine
# why the event was not published.
MaxEventSizeExceeded = Nack.Code.MAX_EVENT_SIZE_EXCEEDED
TopicUnknown = Nack.Code.TOPIC_UNKNOWN
TopicArchived = Nack.Code.TOPIC_ARCHIVED
TopicDeleted = Nack.Code.TOPIC_DELETED
PermissionDenied = Nack.Code.PERMISSION_DENIED
ConsensusFailure = Nack.Code.CONSENSUS_FAILURE
ShardingFailure = Nack.Code.SHARDING_FAILURE
Redirect = Nack.Code.REDIRECT
Internal = Nack.Code.INTERNAL



class ServerStatus(object):
    def __init__(self, status, version, uptime):
        # TODO: convert status enum integer into string
        # TODO: convert uptime into a timedelta
        # TODO: handle the other fields returned from client.status
        self.status = status
        self.version = version
        self.uptime = uptime

    def __str__(self):
        return f"status: {self.status}\nversion: {self.version}\nuptime: {self.uptime}"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/v1beta1/groups.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x61pi/v1beta1/groups.proto\x12\x0e\x65nsign.v1beta1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\"\x91\x03\n\rConsumerGroup\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\x12\n\nproject_id\x18\x02 \x01(\x0c\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x32\n\x08\x64\x65livery\x18\x04 \x01(\x0e\x32 .ensign.v1beta1.DeliverySemantic\x12\x33\n\x10\x64\x65livery_timeout\x18\x05 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x46\n\rtopic_offsets\x18\x0c \x03(\x0b\x32/.ensign.v1beta1.ConsumerGroup.TopicOffsetsEntry\x12\x11\n\tconsumers\x18\r \x03(\x0c\x12+\n\x07\x63reated\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08modified\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a\x33\n\x11TopicOffsetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01*Z\n\x10\x44\x65liverySemantic\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x10\n\x0c\x41T_MOST_ONCE\x10\x01\x12\x11\n\rAT_LEAST_ONCE\x10\x02\x12\x10\n\x0c\x45XACTLY_ONCE\x10\x03\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'api.v1beta1.groups_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CONSUMERGROUP_TOPICOFFSETSENTRY._options = None
  _CONSUMERGROUP_TOPICOFFSETSENTRY._serialized_options = b'8\001'
  _DELIVERYSEMANTIC._serialized_start=513
  _DELIVERYSEMANTIC._serialized_end=603
  _CONSUMERGROUP._serialized_start=110
  _CONSUMERGROUP._serialized_end=511
  _CONSUMERGROUP_TOPICOFFSETSENTRY._serialized_start=460
  _CONSUMERGROUP_TOPICOFFSETSENTRY._serialized_end=511
# @@protoc_insertion_point(module_scope)



import os
import json
import inspect

from ulid import ULID

from pyensign.connection import Client
from pyensign.events import from_object
from pyensign.status import ServerStatus
from pyensign.utils.topics import Topic, TopicCache
from pyensign.connection import Connection
from pyensign.api.v1beta1 import topic_pb2, query_pb2
from pyensign.auth.client import AuthClient
from pyensign.exceptions import (
    CacheMissError,
    UnknownTopicError,
    InvalidQueryError,
    EnsignInvalidArgument,
    EnsignTopicCreateError,
    EnsignTopicNotFoundError,
)


class Ensign:
    """
    Ensign connects to an Ensign server and provides access to the Ensign API.
    """

    def __init__(
        self,
        client_id="",
        client_secret="",
        cred_path="",
        endpoint="ensign.rotational.app:443",
        insecure=False,
        auth_url="https://auth.rotational.app",
        disable_topic_cache=False,
    ):
        """
        Create a new Ensign client with API credentials.

        Parameters
        ----------
        client_id : str
            The client ID part of the API key. If not provided, the client ID is loaded
            from the ENSIGN_CLIENT_ID environment variable.
        client_secret : str
            The client secret part of the API key. If not provided, the client secret
            is loaded from the ENSIGN_CLIENT_SECRET environment variable.
        cred_path : str (optional)
            Load a JSON file containing the 'ClientID' and 'ClientSecret' key.
        endpoint : str (optional)
            The endpoint of the Ensign server.
        insecure : bool (optional)
            Set to True to use an insecure connection.
        auth_url : str (optional)
            The URL of the Ensign authentication server.
        disable_topic_cache: bool (optional)
            Set to True to disable topic ID caching.
        """

        if cred_path:
            try:
                with open(cred_path, "r") as file:
                    data = json.load(file)
                    client_id = data["ClientID"]
                    client_secret = data["ClientSecret"]
            except FileNotFoundError as e:
                raise ValueError("Credentials file not found", str(e))
            except UnicodeDecodeError as e:
                raise ValueError("Error decoding credentials file", str(e))
            except json.JSONDecodeError as e:
                raise ValueError("Credentials file has invalid JSON", str(e))
            except IOError as e:
                raise ValueError("IO error while reading credentials file:", str(e))

        if not client_id or client_id == "":
            client_id = os.environ.get("ENSIGN_CLIENT_ID")
        if not client_secret or client_secret == "":
            client_secret = os.environ.get("ENSIGN_CLIENT_SECRET")
        if client_id is None:
            raise ValueError(
                "client_id is required but not provided or set in the ENSIGN_CLIENT_ID environment variable"
            )
        if client_secret is None:
            raise ValueError(
                "client_secret is required but not provided or set in the ENSIGN_CLIENT_SECRET environment variable"
            )
        if not isinstance(client_id, str):
            raise TypeError("client_id must be a string")
        if not isinstance(client_secret, str):
            raise TypeError("client_secret must be a string")

        creds = {"client_id": client_id, "client_secret": client_secret}
        auth = AuthClient(auth_url, creds)
        connection = Connection(addrport=endpoint, insecure=insecure, auth=auth)

        self.topics = TopicCache(read_only=disable_topic_cache)
        self.client = Client(connection, topic_cache=self.topics)

    async def publish(
        self, topic, *events, on_ack=None, on_nack=None, ensure_exists=False
    ):
        """
        Publish events to an Ensign topic. This function is asynchronous; it publishes
        events to an outgoing queue and returns immediately. Publishers can configure
        the `on_ack` and `on_nack` callbacks to implement per-event success or failure
        handling. Alternatively, publishers can check the status of events at any time
        using the `acked()` and `nacked()` methods on the events themselves.

        Parameters
        ----------
        topic: str
            The name or ID of the topic to publish events to.

        events : iterable of events.Event objects
            The events to publish.

        on_ack: coroutine (optional)
            A coroutine to be invoked when an ACK message is received from Ensign,
            indicating that an event was successfully published. The first argument of
            the coroutine is an Ack object with a `committed` timestamp that indicates
            when the event was committed.

        on_nack: coroutine (optional)
            A coroutine to be invoked when a NACK message is received from Ensign,
            indicating that the event could not be published. The first argument of the
            coroutine is a Nack object with a `code` field that indicates the reason
            why the event was not published.

        ensure_exists: bool (optional)
            Create the topic if it does not exist. With this option, a topic name must
            be provided, not an ID.

        Raises
        ------
        ValueError
            If a topic or no events are provided.

        TypeError
            If the topic is not a string.

        UnknownTopicError
            If the topic does not exist and ensure_exists is False.
        """

        if topic == "":
            raise ValueError("topic is required")

        if not isinstance(topic, str):
            raise TypeError(
                "expected type 'str' for topic, got '{}'".format(type(topic).__name__)
            )

        # Handle a list of events provided as a single argument
        if len(events) == 1 and (
            isinstance(events[0], list) or isinstance(events[0], tuple)
        ):
            events = events[0]

        if len(events) == 0:
            raise ValueError("no events provided")

        if ensure_exists:
            # Get or create the topic if it doesn't exist
            id = ULID.from_str(await self.ensure_topic_exists(topic))
            topic = Topic(id=id, name=topic)
        else:
            try:
                # Try to resolve the topic ID from the string
                topic = Topic(id=self.topics.resolve(topic))
            except CacheMissError:
                # Assume the topic string is a name, the publisher will make a best
                # effort to resolve the ID from the name using the response from the
                # server
                topic = Topic(name=topic)

        # TODO: Support user-defined generators
        async def next():
            for event in events:
                yield event

        await self.client.publish(
            topic,
            next(),
            on_ack=on_ack,
            on_nack=on_nack,
        )

    async def subscribe(self, *topics, query="", consumer_group=None):
        """
        Subscribe to events from the Ensign server. This method returns an async
        generator that yields Event objects, so the `async for` syntax can be used to
        process events as they are received.

        Parameters
        ----------
        topics : iterable of str
            The topic names or IDs to subscribe to.

        query : str (optional)
            EnSQL query to filter events.

        consumer_group : api.v1beta1.groups.ConsumerGroup (optional)
            The consumer group to use for the subscriber.

        Raises
        ------
        ValueError
            If no topics are provided.

        UnknownTopicError
            If a topic is provided that does not exist.

        Yields
        ------
        events.Event
            The event objects. Subscribers should call ack() or nack() on each event
            object to signal to the server if the event was successfully processed or
            it needs to be redelivered to another subscriber in the consumer group.
        """

        # Handle a list of topics passed as a single argument
        if len(topics) == 1 and (
            isinstance(topics[0], list) or isinstance(topics[0], tuple)
        ):
            topics = topics[0]

        if len(topics) == 0:
            raise ValueError("no topics provided")

        # Parse topic names into ID strings
        topic_ids = []
        for topic in topics:
            if not isinstance(topic, str):
                raise TypeError(
                    "expected type 'str' for topic, got {}".format(type(topic))
                )

            # Attempt to lookup or parse the ID
            try:
                topic_ids.append(str(self.topics.resolve(topic)))
            except CacheMissError:
                # Ignore unknown names for now, the server will decide if they are
                # valid
                topic_ids.append(topic)

        # Yield events from the subscribe stream
        async for event in self.client.subscribe(
            topic_ids,
            query=query,
            consumer_group=consumer_group,
        ):
            yield event

    async def query(self, query, params=None):
        """
        Execute an EnSQL query. This method returns a cursor that can be used to fetch
        the results of the query, via `fetchone()`, `fetchmany()`, or `fetchall()`
        methods.

        Parameters
        ----------
        query : str
            A valid EnSQL query. EnSQL queries are similar to SQL queries on relational
            databases, but instead of returning rows in tables they return events in
            topic streams.

            For example, the following query returns the first 10 events in a topic:

            SELECT * FROM <topic> LIMIT 10

            See https://ensign.rotational.dev/ensql for more information on the EnSQL syntax.

        params : dict (optional)
            Parameters to be substituted into the query. Only primitive types are
            supported, i.e. strings, integers, floats, and booleans.

        Returns
        -------
        connection.Cursor
            The cursor object that can be used to fetch the results of the query.

        Raises
        ------
        InvalidQueryError
            If the provided query is invalid.

        QueryNoRows
            If the query returned no results.

        TypeError
            If an unsupported parameter type is provided.
        """

        # Parse the args into the protobuf parameters
        parameters = []
        if params:
            for name, value in params.items():
                if isinstance(value, int):
                    parameters.append(query_pb2.Parameter(name=name, i=value))
                elif isinstance(value, float):
                    parameters.append(query_pb2.Parameter(name=name, d=value))
                elif isinstance(value, bool):
                    parameters.append(query_pb2.Parameter(name=name, b=value))
                elif isinstance(value, bytes):
                    parameters.append(query_pb2.Parameter(name=name, y=value))
                elif isinstance(value, str):
                    parameters.append(query_pb2.Parameter(name=name, s=value))
                else:
                    raise TypeError(
                        "unsupported parameter type: {}".format(type(value).__name__)
                    )

        # Execute the query and return the cursor to the user
        try:
            cursor = await self.client.en_sql(query, params=parameters)
        except EnsignInvalidArgument as e:
            raise InvalidQueryError(e.details)

        return cursor

    async def explain_query(self, query, params):
        raise NotImplementedError

    async def get_topics(self):
        """
        Get all topics.

        Yields
        ------
        api.v1beta1.topic_pb2.Topic
            The topics.
        """

        topics = []
        page = None
        token = ""
        while page is None or token != "":
            page, token = await self.client.list_topics(next_page_token=token)
            topics.extend(page)
        return topics

    async def create_topic(self, topic_name):
        """
        Create a topic.

        Parameters
        ----------
        topic_name : str
            The name for the new topic, must be unique to the project.

        Returns
        -------
        str
            The ID of the topic that was created.
        """

        created = await self.client.create_topic(topic_pb2.Topic(name=topic_name))
        if not created:
            # TODO: Return more specific errors
            raise EnsignTopicCreateError("topic creation failed")
        id = ULID(created.id)
        if self.topics:
            self.topics.add(created.name, id)
        return str(id)

    async def ensure_topic_exists(self, topic_name):
        """
        Check if a topic exists and create it if it does not. This is a shortcut for
        topic_exists and create_topic but also returns the topic ID.

        Parameters
        ----------
        topic_name : str
            The name of the topic to check.

        Returns
        -------
        str
            The ID of the topic.
        """

        if await self.topic_exists(topic_name):
            return await self.topic_id(topic_name)
        else:
            return await self.create_topic(topic_name)

    async def retrieve_topic(self, id):
        raise NotImplementedError

    async def archive_topic(self, id):
        raise NotImplementedError

    async def destroy_topic(self, id):
        """
        Completely destroy a topic including all of its data. This operation is
        asynchronous so it may take some time for the topic to be destroyed.

        Caution! This operation is irreversible.

        Parameters
        ----------
        id : str
            The ID of the topic to destroy.

        Returns
        -------
        bool
            True if the request was accepted, False otherwise.
        """

        _, state = await self.client.destroy_topic(id)
        return state == topic_pb2.TopicTombstone.Status.DELETING

    async def topic_names(self):
        raise NotImplementedError

    async def topic_id(self, name):
        """
        Get the ID of a topic by name.

        Parameters
        ----------
        name: str
            The name of the topic.

        Returns
        -------
        str
            The ID of the topic.

        Raises
        ------
        EnsignTopicNotFoundError
        """

        # Attempt to get the topic ID from the cache
        if self.topics is not None:
            try:
                return str(self.topics.get(name))
            except CacheMissError:
                pass

        # Get the topic ID from Ensign
        page = None
        token = ""
        while page is None or token != "":
            # TODO: We could use topic_names but it currently requires hashing the
            # topic name with murmur3.
            page, token = await self.client.list_topics(next_page_token=token)
            for topic in page:
                if topic.name == name:
                    id = ULID(topic.id)
                    if self.topics is not None:
                        self.topics.add(name, id)
                    return str(id)
        raise EnsignTopicNotFoundError(f"topic not found by name: {name}")

    async def topic_exists(self, name):
        """
        Check if a topic exists by name.

        Parameters
        ----------
        name : str
            The name of the topic to check.

        Returns
        -------
        bool
            True if the topic exists, False otherwise.
        """

        # Attempt to check existence using the cache
        if self.topics is not None and self.topics.exists(name):
            return True

        # Check existence using Ensign
        _, exists = await self.client.topic_exists(topic_name=name)
        return exists

    async def info(self, topic_ids=[]):
        """
        Get aggregated statistics for topics in the project.

        Parameters
        ----------
        topic_ids: list of str (optional)
            If provided, only aggregate info for the specified topic IDs. Otherwise,
            the info includes all of the topics in the project.

        Returns
        -------
        api.v1beta1.ProjectInfo
            The aggregated statistics for the topics in the project.
        """

        if not isinstance(topic_ids, list):
            raise TypeError(f"expected list of topic IDs, got {type(topic_ids)}")

        # Ensure that only topic IDs are provided
        topics = []
        for id in topic_ids:
            try:
                topics.append(ULID.from_str(id).bytes)
            except ValueError:
                raise ValueError(f"not parseable as a topic ID: {id}")

        return await self.client.info(topics)

    async def status(self):
        """
        Check the status of the Ensign server.

        Returns
        -------
        ServerStatus:
            Status of the server.
        """

        status, version, uptime, _, _ = await self.client.status()
        return ServerStatus(status, version, uptime)

    async def close(self):
        """
        Close the Ensign client.
        """
        await self.client.close()

    async def __aenter__(self):
        await self.client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    def _resolve_topic(self, topic):
        """
        Resolve a topic string into a ULID by looking it up in the cache, otherwise
        assume it's a topic ID and try to parse it.
        """

        # Attempt to retrieve the topic ID from the cache
        if self.topics is not None:
            try:
                return self.topics.get(topic)
            except CacheMissError:
                pass

        # Otherwise attempt to parse as a ULID
        try:
            return ULID.from_str(topic)
        except ValueError:
            # TODO: Might need to return the name of the project here
            raise UnknownTopicError(
                f"unknown topic '{topic}', please provide the name or ID of a topic in your project"
            )


##########################################################################
## Decorators
##########################################################################

# Global Ensign client. This is not thread-safe but it's safe to use concurrently in
# asyncio coroutines.
_client = None


def authenticate(*auth_args, **auth_kwargs):
    """
    Decorator function to authenticate with Ensign. This function ideally should only
    be called once, usually at the entry point in the application. Duplicate calls to
    the decorator are ignored.

    By default credentials are loaded from the ENSIGN_CLIENT_ID and ENSIGN_CLIENT_SECRET
    environment variables. Alternatively, credentials can be provided as keyword
    arguments or a JSON file path, see the Ensign class for more details.

    Usage
    -----
    ```
    # Use credentials from the environment
    @authenticate()
    async def main():
        ...

    # Use credentials from kwargs
    @authenticate(client_id="...", client_secret="...")
    async def main():
        ...

    # Use credentials from a JSON file
    @authenticate(cred_path="...")
    async def main():
        ...
    ```
    """

    def wrap_coroutine(coro):
        async def wrapper(*args, **kwargs):
            global _client
            if _client is not None:
                return await coro(*args, **kwargs)

            # Set the global client for the duration of the coroutine
            try:
                async with Ensign(*auth_args, **auth_kwargs) as client:
                    _client = client
                    res = await coro(*args, **kwargs)
            except Exception as e:
                _client = None
                raise e

            _client = None
            return res

        return wrapper

    def wrap_async_generator(coro):
        async def wrapper(*args, **kwargs):
            global _client
            if _client is not None:
                async for res in coro(*args, **kwargs):
                    yield res

            # Set the global client for the duration of the generator
            try:
                async with Ensign(*auth_args, **auth_kwargs) as client:
                    _client = client
                    async for res in coro(*args, **kwargs):
                        yield res
            except Exception as e:
                _client = None
                raise e

            _client = None

        return wrapper

    # Return either a coroutine or an async generator wrapper to match the marked
    # function
    def decorator(coro):
        if inspect.iscoroutinefunction(coro):
            return wrap_coroutine(coro)
        elif inspect.isasyncgenfunction(coro):
            return wrap_async_generator(coro)
        else:
            raise TypeError(
                "decorated function must be a coroutine or async generator, got {}".format(
                    type(coro)
                )
            )

    return decorator


def publish(topic, mimetype=None, encoder=None):
    """
    Decorator to mark a publish function. The return value of the function will be
    published as an event to the topic. If the function is a generator, the events are
    published as they are yielded from the generator. This method assumes that you have
    already authenticated with Ensign using the `@authenticate` decorator.

    Parameters
    ----------
    topic: str
        The name or ID of the topic to publish events to.

    mimetype: str or int (optional)
        The mimetype indicating how the return value will be encoded. If not provided,
        the mimetype is inferred from the return value of the function.

    encoder: events.BytesEncoder (optional)
        The encoder to use to encode the return value into bytes for the event payload.
        If not provided, a best effort is made to encode the value, for most data types
        this will be JSON.

    Usage
    -----
    ```
    # Publish a JSON event to the topic 'my-topic'
    @publish("my-topic")
    async def my_function():
        return {"foo": "bar"}

    # Publish multiple events using a generator
    @publish("my-topic")
    async def my_generator():
        for i in range(3):
            yield {"event_id": i}
    ```
    """

    def wrap_coroutine(coro):
        async def wrapper(*args, **kwargs):
            # Create the Ensign client if not already created
            global _client
            if _client is None:
                raise RuntimeError(
                    "publish requires a connection to Ensign, please use the authenticate decorator to provide your credentials"
                )

            # Call the function and get the return value
            val = await coro(*args, **kwargs)
            if val is None:
                return val

            # Create the event to publish
            event = from_object(val, mimetype=mimetype, encoder=encoder)

            # Publish the event
            await _client.publish(topic, event)

            return val

        return wrapper

    def wrap_async_generator(coro):
        async def wrapper(*args, **kwargs):
            # Create the Ensign client if not already created
            global _client
            if _client is None:
                raise RuntimeError(
                    "publish requires a connection to Ensign, please use the authenticate decorator to provide your credentials"
                )

            # Yield the values from the generator
            async for val in coro(*args, **kwargs):
                if val is None:
                    yield val
                    continue

                # Create the event to publish
                event = from_object(val, mimetype=mimetype, encoder=encoder)

                # Publish the event
                await _client.publish(topic, event)

                yield val

        return wrapper

    # Return either a coroutine or an async generator wrapper to match the marked
    # function
    def decorator(coro):
        if inspect.iscoroutinefunction(coro):
            return wrap_coroutine(coro)
        elif inspect.isasyncgenfunction(coro):
            return wrap_async_generator(coro)
        else:
            raise TypeError(
                "decorated function must be a coroutine or async generator, got {}".format(
                    type(coro)
                )
            )

    return decorator



import grpc
import asyncio

from pyensign.exceptions import EnsignTypeError


class RequestIterator:
    """
    RequestIterator is an asynchronous iterator that yields requests from an internal
    client stream after sending an initial handshaking request. This iterator can be
    used directly as the request iterator for the publish or subscribe RPCs.
    """

    def __init__(self, queue, init_request):
        self.queue = queue
        self.init_request = init_request

    async def __aiter__(self):
        # Send the initial request to the server
        yield self.init_request

        # Publish events from the client until closed
        # When this iterator is done, the gRPC stream will also be closed
        while True:
            req = await self.queue.read_request()
            if req is None:
                break
            yield req


class ResponseIterator:
    """
    ResponseIterator is an asynchronous iterator that reads responses from a gRPC
    stream.
    """

    def __init__(self, stream):
        self.stream = stream
        pass

    async def consume(self):
        async for _ in self:
            pass

    async def __aiter__(self):
        while True:
            try:
                rep = await self.stream.read()
            except grpc.aio.AioRpcError:
                break
            # Handle unexpected end of stream
            if rep is grpc.aio.EOF:
                break
            yield rep


class PublishResponseIterator(ResponseIterator):
    """
    PublishResponseIterator is an asynchronous iterator that reads responses from a
    gRPC publish stream and executes user-defined callbacks for acks and nacks.
    """

    def __init__(self, stream, pending, on_ack=None, on_nack=None):
        """
        Parameters
        ----------
        stream : grpc.aio.StreamStreamCall
            The gRPC stream object which has been opened and is ready to be read from.
        pending : dict
            A dictionary of pending local ULIDs to events that have been published but
            not yet acked or nacked.
        """
        self.stream = stream
        self.pending = pending
        self.on_ack = on_ack
        self.on_nack = on_nack

    async def consume(self):
        """
        Consume all responses from the stream until closed.
        """
        async for rep in self:
            # Handle messages from the server
            rep_type = rep.WhichOneof("embed")
            if rep_type == "ack":
                event = self.pending.pop(rep.ack.id, None)
                if event:
                    event.mark_acked(rep.ack)
                    if self.on_ack:
                        await self.on_ack(rep.ack)
            elif rep_type == "nack":
                event = self.pending.pop(rep.nack.id, None)
                if event:
                    event.mark_nacked(rep.nack)
                    if self.on_nack:
                        await self.on_nack(rep.nack)
            elif rep_type == "close_stream":
                break
            else:
                raise EnsignTypeError(f"unexpected response type: {rep_type}")


class SubscribeResponseIterator(ResponseIterator):
    """
    SubscribeResponseIterator is an asynchronous iterator that reads responses from a
    gRPC subscribe stream and writes events to a client queue.
    """

    def __init__(self, stream, queue):
        self.stream = stream
        self.queue = queue

    async def consume(self):
        """
        Consume all responses from the stream until closed.
        """
        async for rep in self:
            rep_type = rep.WhichOneof("embed")
            if rep_type == "event":
                await self.queue.write_response(rep.event)
            elif rep_type == "close_stream":
                break
            else:
                await self.queue.write_response(
                    EnsignTypeError(f"unexpected response type: {rep_type}")
                )
                break



import re
import json
import time
import pickle
import asyncio
from enum import Enum
from google.protobuf.timestamp_pb2 import Timestamp

from pyensign.ack import Ack
from pyensign.utils import pbtime
from pyensign import mimetypes as mtype
from pyensign.api.v1beta1 import ensign_pb2, event_pb2
from pyensign.exceptions import CouldNotAck, CouldNotNack, NackError


class Event:
    """
    Event is an abstraction of an Ensign protocol buffer Event to make it easier to
    create and parse events.
    """

    def __init__(
        self,
        data=None,
        mimetype=mtype.ApplicationJSON,
        schema_name="Unknown",
        schema_version="0.0.0",
        meta=None,
    ):
        """
        Create a new Event from a mimetype and data.

        Parameters
        ----------
        data : bytes
            The data to use for the event.
        mimetype: str or int
            The mimetype of the data (e.g. "application/json").
        schema_name: str (optional, default: "Unknown")
            The name of the user-defined schema that the event data conforms to.
            Subscribers consuming the event can use the schema to validate and parse
            the event data.
        schema_version: str (optional, default: "0.0.0")
            The semantic version string of the user-defined schema.
        meta: dict (optional)
            A set of key-value pairs to associate with the event.
        """

        if not data:
            raise ValueError("no data provided")

        if mimetype is None:
            raise ValueError("no mimetype provided")

        schema_name = schema_name.strip()
        if schema_name == "":
            raise ValueError("schema name cannot be empty")

        self.mimetype = mtype.parse(mimetype)

        # Fields that the user may want to modify after creation.
        self.data = data
        self.meta = meta
        self.type = Type(
            name=schema_name,
            version=schema_version,
        )
        self.created = Timestamp(seconds=int(time.time()))

        # Ensure that the protocol buffer representation is valid at the point of
        # creation.
        self._proto = self.proto()

        self.id = None
        self.committed = None
        self.error = None
        self._stream = None
        self._state = EventState.INITIALIZED

    def proto(self):
        """
        Return the protocol buffer representation of the event.

        Returns
        -------
        api.v1beta1.event_pb2.Event
            The protocol buffer representation of the event.
        """

        # TODO: Should we raise type errors and missing field errors to help the user?
        return event_pb2.Event(
            data=self.data,
            metadata=self.meta,
            mimetype=self.mimetype,
            type=self.type.proto(),
            created=self.created,
        )

    def acked(self):
        """
        Returns True if the event has been acked.
        """

        return self._state == EventState.ACKED

    def nacked(self):
        """
        Returns True if the event has been nacked.
        """

        return self._state == EventState.NACKED

    def nack_error(self):
        """
        Return the error if the event was nacked.
        """

        return self.error

    async def ack(self):
        """
        Send an acknowledgement back to the Ensign server to indicate that the Event
        has been successfully consumed. For consumer groups that have exactly-once or
        at-least-once delivery guarantees, this signals that the message has been
        delivered successfully to avoid redelivery of the message to another consumer.

        Returns
        -------
        True
            If the Event was already acked.
        False
            If a nack was sent before the ack.

        Raises
        ------
        CouldNotAck
            If the Event was not received by a subscriber.
        """

        if self._state == EventState.ACKED:
            return True
        elif self._state == EventState.NACKED:
            return False
        elif (
            self._state == EventState.INITIALIZED or self._state == EventState.PUBLISHED
        ):
            raise CouldNotAck("Event has not been received by a subscriber")

        await self._stream.write_request(
            ensign_pb2.SubscribeRequest(ack=ensign_pb2.Ack(id=self.id))
        )
        self._state = EventState.ACKED
        return True

    async def nack(self, code):
        """
        Send a negative acknowledgement back to the Ensign server to indicate that the
        Event has not been successfully consumed. For consumer groups that have
        exactly-once or at-least-once delivery guarantees, this signals that the
        message should be delivered to another consumer.

        Parameters
        ----------
        code : int
            The error code to send back to the server.

        Returns
        -------
        True
            If the Event was already nacked.
        False
            If an ack was sent before the nack.

        Raises
        ------
        CouldNotNack
            If the Event was not received by a subscriber.
        """

        if self._state == EventState.NACKED:
            return True
        elif self._state == EventState.ACKED:
            return False
        elif (
            self._state == EventState.INITIALIZED or self._state == EventState.PUBLISHED
        ):
            raise CouldNotNack("Event has not been received by a subscriber")

        await self._stream.write_request(
            ensign_pb2.SubscribeRequest(nack=ensign_pb2.Nack(id=self.id, code=code))
        )
        self._state = EventState.NACKED
        return True

    async def wait_for_ack(self):
        """
        Wait for the event to be acked or nacked. If the event was acked this returns
        the ack. If the event was nacked this raises a NackError.
        """

        while not self.acked() and not self.nacked():
            await asyncio.sleep(0.1)

        if self.nacked():
            raise NackError(self.error.code, self.error.error)

        return Ack(self.id, self.committed)

    def mark_published(self):
        self._state = EventState.PUBLISHED

    def mark_acked(self, ack):
        self.id = ack.id
        self.committed = ack.committed
        self._state = EventState.ACKED

    def mark_nacked(self, nack):
        self.error = nack
        self._state = EventState.NACKED

    def mark_subscribed(self, id, ackback_stream):
        self.id = id
        self._stream = ackback_stream
        self._state = EventState.SUBSCRIBED

    def __repr__(self):
        repr = "Event("
        repr += "data={}, ".format(self.data)
        repr += "mimetype={}, ".format(self.mimetype)
        repr += "schema_name={}, ".format(self.type.name)
        repr += "schema_version={}, ".format(self.type.semver())
        if self.meta:
            repr += "meta={}".format(self.meta)
        repr += ")"
        return repr

    def __str__(self):
        s = "Event:"
        if self.id:
            s += "\n\tid: {}".format(self.id)
        data = str(self.data)
        if len(data) > 100:
            data = data[:100] + "..."
        s += "\n\tdata: {}".format(data)
        s += "\n\tmimetype: {}".format(mtype.to_str(self.mimetype))
        s += "\n\tschema: {}".format(self.type)
        if self.meta:
            s += "\n\tmeta: {}".format(self.meta)
        s += "\n\tstate: {}".format(self._state)
        s += "\n\tcreated: {}".format(str(pbtime.to_datetime(self.created)))
        if self.committed:
            s += "\n\tcommitted: {}".format(str(pbtime.to_datetime(self.committed)))
        if self.error:
            s += "\n\terror: {}".format(self.error)
        return s


class Type:
    """
    Type is a high level abstraction over the protocol buffer type that represents an
    event schema.
    """

    def __init__(self, name, version=None):
        """
        Create a new schema Type from the name and semantic version.

        Parameters
        ----------
        name : str
            The name of the schema.

        version : str (optional)
            The semantic version of the schema which must be parseable as a semver
            2.0.0 string (e.g. "1.0.0").
        """

        self.name = name
        if version:
            self.parse_version(version)

    semver_pattern = re.compile(
        r"^v*(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
    )

    def parse_version(self, version):
        """
        Parse a semver 2.0.0 version string into the major, minor, and patch components.
        Pre-release and build metadata are ignored. See https://semver.org/ and
        https://regex101.com/r/Ly7O1x/3/ for more on parsing.
        """

        match = Type.semver_pattern.search(version)
        if not match:
            raise ValueError("unparseable semantic version: {}".format(version))
        results = match.groupdict()
        self.major_version = int(results["major"])
        self.minor_version = int(results["minor"])
        self.patch_version = int(results["patch"])

    def semver(self):
        """
        Return the semver 2.0.0 representation of the version.
        """

        return "{}.{}.{}".format(
            self.major_version, self.minor_version, self.patch_version
        )

    def proto(self):
        """
        Return the protocol buffer representation of the schema type.
        """

        return event_pb2.Type(
            name=self.name,
            major_version=self.major_version,
            minor_version=self.minor_version,
            patch_version=self.patch_version,
        )

    def __repr__(self):
        return "Type(name={}, version={})".format(self.name, self.semver())

    def __str__(self):
        return "{} v{}".format(self.name, self.semver())


class EventState(Enum):
    # Event has been created but not published
    INITIALIZED = 0
    # Event has been published but not acked by the server
    PUBLISHED = 1
    # Event has been received by subscriber but not acked by the user
    SUBSCRIBED = 2
    # Event has been acked by a user or the server
    ACKED = 3
    # Event has been nacked by a user or the server
    NACKED = 4


def from_object(obj, mimetype=None, encoder=None):
    """
    Makes a best effort attempt to convert a Python object into an Event. If an encoder is provided, it will be used to encode the object to bytes. If no encoder
    is provided, the object will be encoded according to its type or mimetype.

    Parameters
    ----------
    obj : object
        The object to convert into an Event.

    mimetype : str or int (optional)
        The mimetype of the object. If not provided, the mimetype will be inferred from
        the object type.

    encoder : class (optional)
        An encoder with an `encode()` method that encodes the object into bytes. If not
        provided, the object will be encoded according to its type or the provided
        mimetype.
    """

    if mimetype:
        mimetype = mtype.parse(mimetype)

    if encoder:
        # Use the provided encoder to encode the object
        if not mimetype:
            raise ValueError("mimetype must be provided if encoder is provided")
        if not hasattr(encoder, "encode"):
            raise ValueError("encoder must have an encode() method")
        data = encoder.encode(obj)
    elif isinstance(obj, bytes):
        data = obj
        if not mimetype:
            mimetype = mtype.ApplicationOctetStream
    elif isinstance(obj, str):
        data = obj.encode("utf-8")
        if not mimetype:
            mimetype = mtype.TextPlain
    elif mimetype == mtype.ApplicationPythonPickle:
        data = pickle.dumps(obj)
    else:
        # Default case: assume the object is JSON serializable
        data = json.dumps(obj).encode("utf-8")
        if not mimetype:
            mimetype = mtype.ApplicationJSON

    return Event(data=data, mimetype=mimetype)


def from_proto(proto):
    """
    Convert a protocol buffer event into an Event.

    Parameters
    ----------
    proto : api.v1beta1.event_pb2.Event
        The protocol buffer event to convert.

    Returns
    -------
    Event
        The converted event.
    """

    event = Event(
        data=proto.data,
        mimetype=proto.mimetype,
        meta=proto.metadata,
    )

    # Convert the schema type from the protocol buffer
    event.type = Type(
        name=proto.type.name,
    )
    event.type.major_version = proto.type.major_version
    event.type.minor_version = proto.type.minor_version
    event.type.patch_version = proto.type.patch_version
    event.created = proto.created
    return event


async def ack_event(event):
    await event.ack()



import grpc
import time
import json
from urllib import request
from urllib.error import HTTPError, URLError

from pyensign.version import user_agent
from pyensign.exceptions import AuthenticationError
from pyensign.auth.tokens import expires_at, not_before

ACCESS_TOKEN = "access_token"
REFRESH_TOKEN = "refresh_token"


class AuthClient(grpc.AuthMetadataPlugin):
    """
    AuthClient provides an interface for token-based HTTP authentication by requesting
    authentication tokens from a server. Authenticate should be called to obtain an
    initial access token. Refresh should be called periodically to obtain a new access
    token by providing the refresh token to ensure the API connection is uninterrupted.
    """

    def __init__(self, url, creds):
        """
        Create AuthClient from the server URL and API credentials.

        Parameters
        ----------
        url : str
            The full URL of the authentication server including the port (e.g. "http://localhost:8080")
        creds : dict[str, str]
            The credentials to use for authentication
        """

        if url.endswith("/"):
            url = url[:-1]
        self.url = url
        self.creds = creds
        self._access_token = ""
        self._refresh_token = ""
        self._last_login = None

    def __call__(self, _, callback):
        """
        gRPC callback to add the access token on each request.
        TODO: Refresh credentials in the background to avoid blocking RPC calls
        """
        meta = []
        exception = None
        try:
            meta.append(self.credentials())
        except AuthenticationError as e:
            exception = e
        callback(meta, exception)

    def credentials(self):
        """
        Returns the list of credentials to use for token-based authentication. This
        uses the tokens that are already available if possible, otherwise this function
        will attempt to request tokens from the authentication server.
        """

        # If tokens are missing or parital, authenticate
        if not self._access_token or not self._refresh_token:
            self.authenticate()

        # Refresh the access token if not valid
        if not self._access_valid():
            if self._refresh_valid():
                # If there is a valid, non-expired refresh token, try to use it to get
                # a new access token
                try:
                    self.refresh()
                except AuthenticationError:
                    # If refresh fails, attempt to re-authenticate using credentials
                    self.authenticate()
            else:
                self.authenticate()

        # Return the credentials
        return ("authorization", "Bearer " + self._access_token)

    def authenticate(self):
        """
        Calls the authenticate endpoint and sets the access and refresh tokens. Raises
        an AuthenticationError if the request fails or the server returns an error.
        """
        self._do(self.url + "/v1/authenticate", self.creds)

    def refresh(self):
        """
        Calls the refresh endpoint and resets the access and refresh tokens based on
        the response. Raises an AuthenticationError if the request fails or the server
        returns an error.
        """
        if not self._refresh_token:
            raise AuthenticationError("No refresh token available")
        body = {"refresh_token": self._refresh_token}
        self._do(self.url + "/v1/refresh", body)

    def _access_valid(self):
        """
        Returns True if the access token is not expired.
        """
        if not self._access_token:
            return False
        return time.time() < expires_at(self._access_token)

    def _refresh_valid(self):
        """
        Returns True if the refresh token has not expired and it is after the not
        before time.
        """
        if not self._refresh_token:
            return False
        now = time.time()
        return now > not_before(self._refresh_token) and now < expires_at(
            self._refresh_token
        )

    def _do(self, url, body):
        """
        Perform an HTTP request and update the access and refresh tokens based on the
        response.
        """
        data = json.dumps(body).encode("utf-8")
        req = request.Request(
            url, data=data, headers={"User-Agent": user_agent()}, method="POST"
        )
        try:
            with request.urlopen(req) as response:
                rep = json.loads(response.read())
                if ACCESS_TOKEN not in rep:
                    raise AuthenticationError("Missing access_token in response")
                self._access_token = rep[ACCESS_TOKEN]
                if REFRESH_TOKEN in rep:
                    self._refresh_token = rep[REFRESH_TOKEN]
        except HTTPError as e:
            raise AuthenticationError(
                "Failed to authenticate with API credentials: {} error returned".format(
                    e.code
                )
            )
        except URLError as e:
            code, msg = e.reason
            raise AuthenticationError(
                "Failed to connect to authentication server: [{}] {}".format(code, msg)
            )



import asyncio


class WorkerPool:
    """
    WorkerPool manages a pool of async workers to limit the number of coroutines in the
    event loop.
    """

    def __init__(self, max_workers=10, max_queue_size=100):
        self.max_workers = max_workers
        self.workers = []
        self.queue = asyncio.Queue(maxsize=max_queue_size)

    async def schedule(self, coro, done_callback=None):
        """
        Schedule a coroutine to be executed by a worker.
        """
        await self.queue.put((coro, done_callback))
        if len(self.workers) < self.max_workers:
            worker = asyncio.create_task(self._worker())
            self.workers.append(worker)

    async def _worker(self):
        while True:
            try:
                coro, done_callback = self.queue.get_nowait()
                await coro
                if done_callback:
                    done_callback()
            except (asyncio.QueueEmpty, asyncio.CancelledError):
                # If there is nothing to be done or someone cancelled us, stop the
                # worker
                break
        self.workers.remove(asyncio.current_task())

    async def release(self):
        await asyncio.gather(*self.workers)



import asyncio


class BidiQueue:
    """
    BidiQueue implements an in-memory bidirectional buffer for non-blocking
    communication between coroutines.
    """

    def __init__(self, max_queue_size=100):
        self._request_queue = asyncio.Queue(maxsize=max_queue_size)
        self._response_queue = asyncio.Queue(maxsize=max_queue_size)

    async def write_request(self, message):
        """
        Write a request to the queue, blocks if the queue is full.
        """
        await self._request_queue.put(message)

    async def read_request(self):
        """
        Read a request from the queue, blocks if the queue is empty.
        """
        return await self._request_queue.get()

    async def write_response(self, message):
        """
        Write a response to the queue, blocks if the queue is full.
        """
        await self._response_queue.put(message)

    async def read_response(self):
        """
        Read a response from the queue, blocks if the queue is empty.
        """
        return await self._response_queue.get()

    async def close(self):
        """
        Close the queue by sending a message to both the request and response queues.
        The main purpose of this is to unblock pending consumers. Otherwise, consumers
        will block indefinitely on read_request() or read_response(). This means that
        consumers should be prepared to handle None responses from reads.
        """
        await self._request_queue.put(None)
        await self._response_queue.put(None)



import jwt


def expires_at(token):
    """
    Get the expiration unix timestamp of a JWT token.
    """
    return jwt.decode(token, options={"verify_signature": False})["exp"]


def not_before(token):
    """
    Get the not before unix timestamp of a JWT token.
    """
    return jwt.decode(token, options={"verify_signature": False})["nbf"]



from ulid import ULID

from pyensign.utils.cache import Cache
from pyensign.exceptions import CacheMissError


class Topic:
    """
    Topics have a user-defined name but are also unique by ULID. This class stores both
    representations to make topics easier to work with.
    """

    def __init__(self, id=None, name="", topic_str=""):
        self.id = id
        self.name = name
        self.topic_str = topic_str

    def __hash__(self):
        if self.id is None:
            raise ValueError("cannot hash topic with no ID")
        return hash(str(self.id))

    def __eq__(self, other):
        return self.id == other.id


class TopicCache(Cache):
    """
    TopicCache extends the functionality of the Cache class to support topic ID parsing.
    """

    def resolve(self, topic_str):
        """
        Resolve a topic string into a ULID by looking it up in the cache, otherwise
        assume it's a topic ID and try to parse it.
        """

        # Attempt to retrieve the topic ID from the cache
        try:
            return self.get(topic_str)
        except CacheMissError:
            pass

        # Otherwise attempt to parse as a ULID
        try:
            return ULID.from_str(topic_str)
        except ValueError:
            raise CacheMissError(topic_str)



import inspect
from grpc import StatusCode
from grpc.aio import AioRpcError
from functools import wraps

##########################################################################
## Decorators
##########################################################################


def _wrap_rpc(fn):
    @wraps(fn)
    def wrap(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            _handle_client_error(e)

    return wrap


def _wrap_async_rpc(coro):
    @wraps(coro)
    async def wrap(*args, **kwargs):
        try:
            return await coro(*args, **kwargs)
        except Exception as e:
            _handle_client_error(e)

    return wrap


def _wrap_generator(fn):
    @wraps(fn)
    def wrap(*args, **kwargs):
        try:
            for rep in fn(*args, **kwargs):
                yield rep
        except Exception as e:
            _handle_client_error(e)

    return wrap


def _wrap_async_generator(coro):
    @wraps(coro)
    async def wrap(*args, **kwargs):
        try:
            async for rep in coro(*args, **kwargs):
                yield rep
        except Exception as e:
            _handle_client_error(e)

    return wrap


def catch_rpc_error(coro):
    """
    Decorate coroutines to catch and handle gRPC client errors.
    """

    if inspect.isgeneratorfunction(coro):
        return _wrap_generator(coro)
    elif inspect.isasyncgenfunction(coro):
        return _wrap_async_generator(coro)
    elif inspect.iscoroutinefunction(coro):
        return _wrap_async_rpc(coro)
    else:
        return _wrap_rpc(coro)


##########################################################################
## Error Handling
##########################################################################


def _handle_client_error(e):
    """
    Provides error handling for errors raised by the gRPC client.
    """

    if isinstance(e, AioRpcError):
        code = e.code()
        args = (code.name, code.value, e.details())
        if code == StatusCode.INVALID_ARGUMENT:
            raise EnsignInvalidArgument(*args) from e
        raise EnsignRPCError(*args) from e
    elif isinstance(e, AttributeError):
        raise EnsignAttributeError(
            "error accessing field from Ensign response: {}".format(e)
        ) from e
    elif isinstance(e, QueryNoRows):
        raise e
    elif isinstance(e, EnsignTopicNotFoundError):
        raise UnknownTopicError(e.topic) from e
    elif isinstance(e, EnsignTypeError):
        raise EnsignTypeError("unexpected type in Ensign response: {}".format(e)) from e
    elif isinstance(e, EnsignInitError):
        raise EnsignInitError(
            "error processing request: client is not connected"
        ) from e
    elif isinstance(e, EnsignTimeoutError):
        raise EnsignTimeoutError(
            "timeout exceeded while connecting to Ensign: {}".format(e)
        ) from e
    elif isinstance(e, EnsignClientClosingError):
        raise EnsignClientClosingError(
            "error processing request: client is already closing"
        ) from e
    else:
        raise PyEnsignError("unknown error: {}".format(e)) from e


##########################################################################
## PyEnsign Exceptions
##########################################################################


class TopicNotFoundError(Exception):
    """
    Raised when a topic is not found
    """

    def __init__(self, topic):
        self.topic = topic

    def __str__(self):
        return "topic not found: {}".format(self.topic)

    pass


class PyEnsignError(Exception):
    """
    Base class for PyEnsign exceptions
    """

    pass


class AuthenticationError(PyEnsignError):
    """
    Raised when PyEnsign fails to authenticate with the Auth server
    """

    pass


class CacheMissError(PyEnsignError):
    """
    Raised when PyEnsign fails to find a cached value
    """

    pass


class UnknownTopicError(PyEnsignError, TopicNotFoundError):
    """
    Raised when PyEnsign fails to parse a topic
    """

    def __str__(self):
        return "unknown topic: {}, please specify the name or ID of a topic in your project".format(
            self.topic
        )

    pass


class CouldNotAck(PyEnsignError):
    """
    Raised when PyEnsign could not ack an event
    """

    pass


class CouldNotNack(PyEnsignError):
    """
    Raised when PyEnsign could not nack an event
    """

    pass


class NackError(PyEnsignError):
    """
    Nack errors returned to the user
    TODO: Handle error codes
    """

    def __init__(self, code, error):
        self.code = code
        self.error = error

    def __str__(self):
        return "event was nacked: {} ({})".format(self.error, self.code)

    pass


class EnSQLError(PyEnsignError):
    """
    Query-related errors returned to the user
    """

    pass


class InvalidQueryError(EnSQLError):
    """
    Raised when a query has invalid syntax
    """

    pass


class QueryNoRows(EnSQLError):
    """
    Raised when a query returns no rows, this means that the query was valid but no
    results were selected by the query
    """

    pass


class CursorError(PyEnsignError):
    """
    Raised when PyEnsign encounters an error from an event cursor
    """

    pass


class CursorNoRows(CursorError):
    """
    Raised when a cursor has no rows (e.g. no results from a query)
    """

    pass


class CursorClosedError(CursorError):
    """
    Raised when a cursor is already closed
    """

    pass


class EnsignError(PyEnsignError):
    """
    Raised when PyEnsign receives an error from the Ensign server
    """

    pass


class EnsignRPCError(EnsignError):
    """
    Raised when PyEnsign receives a gRPC error from the Ensign server
    """

    def __init__(self, name, code, details):
        self.name = name
        self.code = code
        self.details = details

    def __str__(self):
        return "received gRPC error from Ensign: {}: {} ({})".format(
            self.name, self.details, self.code
        )

    pass


class EnsignInvalidArgument(EnsignRPCError):
    """
    Invalid arugment errors returned by Ensign
    """

    pass


class EnsignTypeError(EnsignError, TypeError):
    """
    Raised when PyEnsign receives an unexpected type in a response message
    """

    pass


class EnsignAttributeError(EnsignError, AttributeError):
    """
    Raised when PyEnsign encounters an error trying to access a field from a response message
    """

    pass


class EnsignTopicCreateError(EnsignError):
    """
    Raised when Ensign failed to create a topic
    """

    pass


class EnsignInvalidTopicError(EnsignError, ValueError):
    """
    Raised when a topic does not have all required fields
    """

    pass


class EnsignTopicNotFoundError(EnsignError, TopicNotFoundError):
    """
    Raised when a topic could not be retrieved from Ensign
    """

    def __init__(self, topic):
        self.topic = topic

    def __str__(self):
        return "topic not found: {}".format(self.topic)

    pass


class EnsignInitError(EnsignError):
    """
    Raised when the Ensign client is not initialized
    """

    pass


class EnsignTimeoutError(EnsignError):
    """
    Raised when the Ensign client times out waiting for a response from the server
    """

    pass


class EnsignClientClosingError(EnsignError):
    """
    Raised when the Ensign client is closing but a request is made
    """

    pass



from datetime import datetime


def to_datetime(pbtime):
    """
    Convert a google.protobuf.Timestamp into a datetime.datetime.
    """
    # Temporary workaround for zero-valued timestamps
    if pbtime.seconds < 0:
        return None
    return datetime.fromtimestamp(pbtime.seconds + pbtime.nanos / 1e9)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mimetype/v1beta1/mimetype.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fmimetype/v1beta1/mimetype.proto\x12\x10mimetype.v1beta1*\xdc\x05\n\x04MIME\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0e\n\nTEXT_PLAIN\x10\x01\x12\x0c\n\x08TEXT_CSV\x10\x02\x12\r\n\tTEXT_HTML\x10\x03\x12\x11\n\rTEXT_CALENDAR\x10\x04\x12\x1c\n\x18\x41PPLICATION_OCTET_STREAM\x10\x00\x12\x14\n\x10\x41PPLICATION_JSON\x10\x32\x12\x19\n\x15\x41PPLICATION_JSON_UTF8\x10\x32\x12\x17\n\x13\x41PPLICATION_JSON_LD\x10\x33\x12\x1a\n\x16\x41PPLICATION_JSON_LINES\x10\x34\x12\x16\n\x12\x41PPLICATION_UBJSON\x10\x35\x12\x14\n\x10\x41PPLICATION_BSON\x10\x36\x12\x13\n\x0f\x41PPLICATION_XML\x10\x64\x12\x14\n\x10\x41PPLICATION_ATOM\x10\x65\x12\x18\n\x13\x41PPLICATION_MSGPACK\x10\xfc\x01\x12\x18\n\x13\x41PPLICATION_PARQUET\x10\xfd\x01\x12\x15\n\x10\x41PPLICATION_AVRO\x10\xfe\x01\x12\x19\n\x14\x41PPLICATION_PROTOBUF\x10\xff\x01\x12\x14\n\x0f\x41PPLICATION_PDF\x10\x80\x04\x12\x1d\n\x18\x41PPLICATION_JAVA_ARCHIVE\x10\x81\x04\x12\x1e\n\x19\x41PPLICATION_PYTHON_PICKLE\x10\x82\x04\x12\x14\n\x0fUSER_SPECIFIED0\x10\xe8\x07\x12\x14\n\x0fUSER_SPECIFIED1\x10\xe9\x07\x12\x14\n\x0fUSER_SPECIFIED2\x10\xea\x07\x12\x14\n\x0fUSER_SPECIFIED3\x10\xeb\x07\x12\x14\n\x0fUSER_SPECIFIED4\x10\xec\x07\x12\x14\n\x0fUSER_SPECIFIED5\x10\xed\x07\x12\x14\n\x0fUSER_SPECIFIED6\x10\xee\x07\x12\x14\n\x0fUSER_SPECIFIED7\x10\xef\x07\x12\x14\n\x0fUSER_SPECIFIED8\x10\xf0\x07\x12\x14\n\x0fUSER_SPECIFIED9\x10\xf1\x07\x1a\x02\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mimetype.v1beta1.mimetype_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MIME._options = None
  _MIME._serialized_options = b'\020\001'
  _MIME._serialized_start=54
  _MIME._serialized_end=786
# @@protoc_insertion_point(module_scope)
